from argparse import Namespace
from datetime import datetime
from re import X
from flask import flash, Flask, redirect, jsonify, render_template, request, url_for, g
from flask_sqlalchemy import SQLAlchemy
from sqlalchemy import func, and_, select, delete, true
from sqlalchemy.orm import Session
from flask_restx import Api, Resource, fields
from flask_socketio import SocketIO, emit
from werkzeug.middleware.proxy_fix import ProxyFix
from time import sleep
from threading import Thread
import queue
import logging as log
from logging.config import dictConfig
import os
from subprocess import run
import signal
import json
from sqlalchemy.dialects import sqlite
from .util import PropagatingThread, package_path
from .default_settings import SQLALCHEMY_POOL_SIZE, SQLALCHEMY_DATABASE_URI
from .ansible.scitq.sqlite_inventory import scitq_inventory


MAIN_THREAD_SLEEP = 5
WORKER_OFFLINE_DELAY = 15
SCITQ_SERVER = os.environ.get('SCITQ_SERVER',None)
WORKER_CREATE = f'cd {package_path("ansible","playbooks")} && ansible-playbook deploy_one_vm.yaml --extra-vars "nodename={{hostname}} concurrency={{concurrency}} status=running flavor={{flavor}} region={{region}}"'
if SCITQ_SERVER is not None:
    WORKER_CREATE = WORKER_CREATE[:-1] + f' target={SCITQ_SERVER}"'
WORKER_DELETE = os.environ.get('WORKER_DELETE',f'cd {package_path("ansible","playbooks")} && ansible-playbook destroy_vm.yaml --extra-vars "nodename={{hostname}}"')
WORKER_IDLE_CALLBACK = os.environ.get('WORKER_IDLE_CALLBACK',WORKER_DELETE)
WORKER_CREATE_CONCURRENCY = 10
WORKER_CREATE_RETRY=2
WORKER_CREATE_RETRY_SLEEP=30
UI_OUTPUT_TRUNC=100
WORKER_DESTROY_RETRY=2

dictConfig({
    'version': 1,
    'formatters': {'default': {
        'format': '[%(asctime)s] %(levelname)s in %(module)s: %(message)s',
    }},
    'handlers': {'wsgi': {
        'class': 'logging.StreamHandler',
        'stream': 'ext://flask.logging.wsgi_errors_stream',
        'formatter': 'default'
    }, "file": {
        "class": "logging.handlers.RotatingFileHandler",
        "formatter": "default",
        "filename": os.environ.get('LOG_FILE',"/tmp/scitq.log"),
        "maxBytes": int(os.environ.get('LOG_FILE_MAX_SIZE',"10000000")),
        "backupCount": int(os.environ.get('LOG_FILE_KEEP',"3"))
    }},
    'root': {
        'level': os.environ.get('LOG_LEVEL',"INFO"),
        'handlers': ['wsgi' if 'DEBUG' in os.environ else 'file']
    }
})

IS_SQLITE = 'sqlite' in SQLALCHEMY_DATABASE_URI


log.info('Starting')
log.warning(f'WORKER_CREATE is {WORKER_CREATE}')

worker_create_queue = queue.Queue()


# via https://github.com/pallets/flask-sqlalchemy/blob/main/examples/hello/hello.py
app = Flask(__name__, instance_relative_config=True)
#app.config['SQLALCHEMY_DATABASE_URI'] = "postgresql://postgres:postgres@localhost:5432/"
app.config.from_object('scitq.default_settings')
#app.config.from_pyfile("scitq.cfg", silent=True)
#app.config.from_prefixed_env()
if SQLALCHEMY_POOL_SIZE is not None:
    db = SQLAlchemy(app, engine_options={'pool_size': int(SQLALCHEMY_POOL_SIZE)})
else:
    db = SQLAlchemy(app)
socketio = SocketIO(app)

class Task(db.Model):
    __tablename__ = "task"
    task_id = db.Column(db.Integer, primary_key=True)
    name = db.Column(db.String)
    command = db.Column(db.String, nullable=False)
    status = db.Column(db.String, nullable=False)
    creation_date = db.Column(db.DateTime, nullable=False)
    modification_date = db.Column(db.DateTime)
    batch = db.Column(db.String, nullable=True)
    input = db.Column(db.String, nullable=True)
    output = db.Column(db.String, nullable=True)
    container = db.Column(db.String, nullable=True)
    container_options = db.Column(db.String, nullable=True)
    resource = db.Column(db.String, nullable=True)

    def __init__(self, command, name=None, status='pending', batch=None, 
                    input=None, output=None, container=None, 
                    container_options=None, resource=None):
        self.name = name
        self.command = command
        self.status = status
        self.creation_date = datetime.utcnow()
        self.modification_date = self.creation_date
        self.batch = batch
        self.input = input
        self.output = output
        self.container = container
        self.container_options = container_options
        self.resource = resource


class Worker(db.Model):
    __tablename__ = "worker"
    worker_id = db.Column(db.Integer,primary_key=True)
    name = db.Column(db.String, nullable=False)
    hostname = db.Column(db.String)
    status = db.Column(db.String, nullable=False)
    concurrency = db.Column(db.Integer)
    prefetch = db.Column(db.Integer)
    load = db.Column(db.String)
    memory=db.Column(db.String)
    read_bytes=db.Column(db.String)
    written_bytes=db.Column(db.String)
    creation_date = db.Column(db.DateTime)
    modification_date = db.Column(db.DateTime)
    last_contact_date = db.Column(db.DateTime)
    batch = db.Column(db.String,nullable=True)
    idle_callback = db.Column(db.String, nullable=True)

    def __init__(self, name, concurrency, prefetch=0, hostname=None, 
                status='paused', batch=None, idle_callback=None):
        self.name = name
        self.concurrency = concurrency
        self.prefetch = prefetch
        self.status = status
        self.creation_date = datetime.utcnow()
        self.modification_date = self.creation_date
        self.hostname = hostname
        self.batch = batch
        self.idle_callback = idle_callback
    
    def destroy(self):
        """Destroy self (stop in provider renting)"""
        log.warning(f'Worker {self.name} is to be destroyed.')

        for i in range(WORKER_DESTROY_RETRY):
            try:
                run([self.idle_callback.format(**(self.__dict__))],shell=True,check=True)
            except Exception as e:
                log.exception(e)

            worker_check = json.loads(scitq_inventory(host=self.hostname))

            if not worker_check:
                log.warning(f'Worker {self.name} was destroyed.')
                break
            else:
                log.exception(f'Worker {self.name} was not properly destroyed (f{worker_check}), retrying.')

class Execution(db.Model):
    __tablename__ = "execution"
    execution_id = db.Column(db.Integer, primary_key=True)
    worker_id = db.Column(db.Integer, db.ForeignKey("worker.worker_id"), nullable=True)
    worker = db.relationship(
        Worker,
        backref=db.backref('executions',
                         uselist=True,
                         cascade='save-update',
                         order_by='Execution.modification_date.desc()'))
    task_id = db.Column(db.Integer, db.ForeignKey("task.task_id"), nullable=False)
    task = db.relationship(
        Task,
        backref=db.backref('tasks',
                         uselist=True,
                         cascade='delete,all'),
                         order_by='Execution.creation_date')
    status = db.Column(db.String, nullable=False)
    creation_date = db.Column(db.DateTime, nullable=False)
    modification_date = db.Column(db.DateTime)
    output = db.Column(db.Text)
    error = db.Column(db.Text)
    return_code = db.Column(db.Integer)
    pid = db.Column(db.Integer)
    output_files = db.Column(db.String, nullable=True)
    

    def __init__(self, worker_id, task_id, status='pending', pid=None, 
                    return_code=None):
        self.worker_id = worker_id
        self.task_id = task_id
        self.status = status
        self.pid = pid
        self.return_code = return_code
        self.creation_date = datetime.utcnow()
        self.modification_date = self.creation_date

class Signal(db.Model):
    __tablename__ = "signal"
    execution_id = db.Column(db.Integer, db.ForeignKey("execution.execution_id"), primary_key=True)
    worker_id = db.Column(db.Integer, db.ForeignKey("worker.worker_id"), primary_key=True)
    signal = db.Column(db.Integer, nullable=False)    

    def __init__(self, execution_id, worker_id, signal):
        self.execution_id = execution_id
        self.worker_id = worker_id
        self.signal = signal

with app.app_context():
    db.create_all()


# via https://flask-restx.readthedocs.io/en/latest/example.html
app.wsgi_app = ProxyFix(app.wsgi_app)
api = Api(app, version='1.0', title='TaskMVC API',
    description='A simple TaskMVC API'
)


class BaseDAO(object):
    ObjectType = None
    authorized_status = []

    def get(self, id):
        object = self.ObjectType.query.get(id)
        if object is None:
            api.abort(404, "{} {} doesn't exist".format(
                    self.ObjectType.__name__,id))
        return object
        
    def create(self, data):
        object = self.ObjectType(**data)
        if 'status' in data and data['status'] not in self.authorized_status:
            api.abort(500,
                f"Status {data['status']} is not possible (only {' '.join(self.authorized_status)})")
        db.session.add(object)
        db.session.commit()
        return object

    def update(self, id, data):
        object = self.get(id)
        modified = False
        for attr, value in data.items():
            if hasattr(object,attr): 
                if getattr(object,attr)!=value:
                    if attr=='status' and value not in self.authorized_status:
                        api.abort(500,
                           f"Status {value} is not possible (only {' '.join(self.authorized_status)})")
                    setattr(object, attr, value)
                    modified = True
            else:
                api.abort(500,f'Error: {object.__name__} has no attribute {attr}')
        if modified:
            if hasattr(object,'modification_date'):
                object.modification_date = datetime.utcnow()
            db.session.commit()
        return object

    def delete(self, id):
        object = self.get(id)
        db.session.delete(object)
        db.session.commit()
        return object
    
    def list(self, *args, sorting_column=None):
        if args:
            return list(self.ObjectType.query.filter(*args).order_by(sorting_column))
        else:
            return list(self.ObjectType.query.order_by(sorting_column).all())
        
        



ns = api.namespace('tasks', description='TASK operations')

class TaskDAO(BaseDAO):
    ObjectType = Task
    authorized_status = ['paused','pending','assigned','accepted','running','failed','succeeded']

    def list(self):
        return super().list(sorting_column='task_id')

task_dao = TaskDAO()

task = api.model('Task', {
    'task_id': fields.Integer(readonly=True, description='The task unique identifier'),
    'name': fields.String(required=False, description='The task (optional) name'),
    'status': fields.String(required=False, 
        description=f'The task status: {", ".join(TaskDAO.authorized_status)}'), 
    'command': fields.String(required=True, description='The task command'), 
    'creation_date': fields.DateTime(readonly=True, 
        description='timestamp of task creation in server'),
    'modification_date': fields.DateTime(readonly=True,
        description='timestamp of task last modification'),
    'batch': fields.String(required=False, 
        description="only workers with the same batch (null or not) will accept the task."),
    'input': fields.String(required=False,
        decription="Input data required for task (space separated files URL in s3://...)"),
    'output': fields.String(required=False,
        decription="Output data basepath where /output content will be copied (if null, the result transmission is up to the command)"),
    'container': fields.String(required=False,
        decription="Container (as for now this is a docker) in which task is launched"),
    'container_options': fields.String(required=False,
        decription="Container (extra) option if needed"),
    'resource': fields.String(required=False,
        decription="Resource data required for task (much like input except it is shared between tasks) (space separated files URL in s3://...)"),
})


@ns.route('/')
class TaskList(Resource):
    '''Shows a list of all tasks, and lets you POST to add new tasks'''
    @ns.doc('list_tasks')
    @ns.marshal_list_with(task)
    def get(self):
        '''List all tasks'''
        return task_dao.list()

    @ns.doc('create_task')
    @ns.expect(task)
    @ns.marshal_with(task, code=201)
    def post(self):
        '''Create a new task'''
        return task_dao.create(api.payload), 201


@ns.route("/<id>")
@ns.param("id", "The task identifier")
@ns.response(404, "Task not found")
class WorkerObject(Resource):
    @ns.doc("get_task")
    @ns.marshal_with(task)
    def get(self, id):
        """Fetch a task given its identifier"""
        return task_dao.get(id)

    @ns.doc("update_task")
    @ns.expect(task)
    @ns.marshal_with(task, code=201)
    def put(self, id):
        """Update a task"""
        return task_dao.update(id, api.payload)

    @ns.doc("delete_task")
    @ns.marshal_with(task)
    def delete(self, id):
        """Delete a task"""
        return task_dao.delete(id)

ns = api.namespace('workers', description='WORKER operations')



class WorkerDAO(BaseDAO):
    ObjectType = Worker
    authorized_status = ['paused','running','offline','terminated']

    def update_contact(self, id, load,memory,read_bytes,written_bytes):
        db.engine.execute(
            db.update(Worker
                    ).values({'last_contact_date':datetime.utcnow(),
                        'load':load,'memory':memory,'read_bytes':read_bytes,'written_bytes':written_bytes}
                    ).where(Worker.worker_id==id)
        )
        db.session.commit()

    def delete(self,id):
        """Delete a worker
        """
        worker=self.get(id)
        log.warning(f'Deleting worker {id} ({worker.idle_callback})')
        if worker.idle_callback is not None:
            with app.app_context():
                session = Session(db.engine)
            def _this_worker_destroy():
                worker.destroy()
                session.delete(worker)
                session.commit()
                session.close()
            Thread(target=_this_worker_destroy).start()
            return worker
        else:
            return super().delete(id)

    def list(self):
        return super().list(sorting_column='worker_id')

worker_dao = WorkerDAO()

worker = api.model('Worker', {
    'worker_id': fields.Integer(readonly=True, description='The worker unique identifier'),
    'name': fields.String(required=True, description='The worker name'),
    'concurrency': fields.Integer(required=True, 
        description='The worker concurrency (nb of parallel processes)'), 
    'prefetch': fields.Integer(required=False, 
        description='How many jobs should be prefetch so that they can be launched as soon as possible'),     
    'hostname': fields.String(required=False, description='The worker hostname'),
    'status': fields.String(required=False,
         description=f'The worker status: {", ".join(WorkerDAO.authorized_status)}'), 
    'load': fields.String(readonly=True, description='The worker load (in %)'),
    'memory':fields.Float(readonly=True, description='Memory not use (in %)'),
    'read_bytes':fields.Float(readonly=True,description='read bytes'),
    'written_bytes':fields.Float(readonly=True,description='written bytes'),
    'creation_date': fields.DateTime(readonly=True, 
        description='timestamp of worker creation'),
    'modification_date': fields.DateTime(readonly=True, 
        description='timestamp of last worker modification'),
    'last_contact_date': fields.DateTime(readonly=True, 
        description='timestamp of last worker ping (automatically sent by worker'),
    'batch': fields.String(required=False, 
        description="worker accept only tasks with same batch (null or not)."),
    'idle_callback': fields.String(readonly=True,
        description="A command to be called on scitq server when the worker load *returns* to zero. Typically used to end cloud instances.")
})

@ns.route('/')
class WorkerList(Resource):
    '''Shows a list of all workers, and lets you POST to add new workers'''
    @ns.doc('list_workers')
    @ns.marshal_list_with(worker)
    def get(self):
        '''List all workers'''
        return worker_dao.list()

    @ns.doc('create_worker')
    @ns.expect(worker)
    @ns.marshal_with(worker, code=201)
    def post(self):
        '''Create a new worker'''
        return worker_dao.create(api.payload), 201


worker_tasks = api.model('WorkerTasks', {
    'worker_id': fields.Integer(readonly=True, description='A worker unique identifier'),
    'status': fields.String(readonly=True, description='The different status of tasks'), 
    'count': fields.Integer(readonly=True, description='How many tasks of this type'),
})

@ns.route("/tasks")
class WorkerTaskList(Resource):
    '''Shows a list of all tasks in all workers, and lets you POST to add new workers'''
    @ns.doc('list_worker_tasks')
    @ns.marshal_list_with(worker_tasks)
    def get(self):
        '''List all workers'''
        return list(db.session.execute("""SELECT w.worker_id, e.status, count(e.task_id) as count 
            FROM worker w
            JOIN execution e ON (e.worker_id=w.worker_id)
            JOIN task t ON e.task_id=t.task_id 
                AND e.execution_id=
                    (SELECT max(execution_id) FROM execution e2 WHERE e2.task_id=t.task_id)
            GROUP BY w.worker_id,e.status"""))
    

@ns.route("/<id>")
@ns.param("id", "The worker identifier")
@ns.response(404, "Worker not found")
class WorkerObject(Resource):
    @ns.doc("get_worker")
    @ns.marshal_with(worker)
    def get(self, id):
        """Fetch a worker given its identifier"""
        return worker_dao.get(id)

    @ns.doc("update_worker")
    @ns.expect(worker)
    @ns.marshal_with(worker, code=201)
    def put(self, id):
        """Update a worker"""
        #worker_dao.update_contact(id)
        if 'batch' in api.payload and api.payload['batch']=='':
            api.payload['batch']==None
        return worker_dao.update(id, api.payload)

    @ns.doc("delete_worker")
    @ns.marshal_with(worker)
    def delete(self, id):
        """Delete a worker"""
        return worker_dao.delete(id)

ping_parser = api.parser()
ping_parser.add_argument('load', type=str, help='Worker load', location='json')
ping_parser.add_argument('memory', type=float, help='Worker load', location='json')
ping_parser.add_argument('read_bytes', type=float, help='Worker load', location='json')
ping_parser.add_argument('written_bytes', type=float, help='Worker load', location='json')

@ns.route("/<id>/ping")
@ns.param("id", "The worker identifier")
@ns.response(404, "Worker not found")
class WorkerPing(Resource):
    @ns.doc("update_worker_contact")
    @ns.expect(ping_parser)
    @ns.marshal_with(worker)
    def put(self, id):
        """Update a worker last contact"""
        args = ping_parser.parse_args()
        worker_dao.update_contact(id, args.get('load',''),args.get('memory',''),args.get('read_bytes',''),args.get('written_bytes',''))
        return worker_dao.get(id)

callback_parser = api.parser()
callback_parser.add_argument('message', type=str, help='Callback message sent (idle)', location='json')

@ns.route("/<id>/callback")
@ns.param("id", "The worker identifier")
@ns.response(404, "Worker not found")
class WorkerCallback(Resource):
    @ns.doc("update_worker_contact")
    @ns.expect(callback_parser)
    def put(self, id):
        """Update a worker last contact"""
        message = callback_parser.parse_args().get('message','')
        worker = worker_dao.get(id)
        if message == 'idle' and worker.idle_callback:
            if db.session.query(Execution).filter(Execution.status=='running',
                    Execution.worker_id==worker.worker_id).count()>0:
                log.warning(f'Worker {worker.name} called idle callback but some tasks are still running, refusing...')
                return {'result':'still have running tasks'}
            if db.session.query(Task).filter(and_(Task.status.in_(['pending','accepted']),
                                    Task.batch==worker.batch)).count()>0:
                log.warning(f'Worker {worker.name} called idle but some tasks are still due...')
                return {'result':'still some work to do, lazy one!'}
            log.warning(f'Worker {worker.name} called idle callback, launching: '+worker.idle_callback.format(**(worker.__dict__)))
            worker.destroy()
            db.session.delete(worker)
            db.session.commit()
            return {'result':'ok'}
        else:
            log.warning(f'Worker {worker.name} called idle callback but has no idle command')
            return {'result':'nothing to do'}

deploy_parser = api.parser()
deploy_parser.add_argument('number', type=int, help='How many workers should be deployed', location='json')
deploy_parser.add_argument('region', type=str, help='Which provider region for worker', location='json')
deploy_parser.add_argument('flavor', type=str, help='Which provider flavor/model for worker', location='json')
deploy_parser.add_argument('batch', type=str, help='Batch name (that must be shared by tasks) for worker', location='json')
deploy_parser.add_argument('concurrency', type=int, help='How many tasks should be run in parallel', location='json')
deploy_parser.add_argument('prefetch', type=int, help='How many extra tasks should be prefetched', location='json')

@ns.route("/deploy")
class WorkerDeploy(Resource):
    @ns.doc("deploy_worker_vm_and_process")
    @ns.expect(deploy_parser)
    def put(self):
        """Create and deploy one or several workers"""
        deploy_args = deploy_parser.parse_args()

        for _ in range(deploy_args['number']):
            worker_create_queue.put(Namespace(concurrency=deploy_args['concurrency'], 
                prefetch=deploy_args['prefetch'],
                flavor=deploy_args['flavor'],
                region=deploy_args['region'],
                batch=deploy_args['batch']))
        
        return {'result':'ok'}

class ExecutionDAO(BaseDAO):
    ObjectType = Execution
    authorized_status = ['pending','accepted','running','failed','succeeded']

    def create(self, data):
        task = task_dao.get(data['task_id'])
        if task.status not in ['pending','paused']:
            api.abort(500, f"A new execution for task {task.task_id} is not possible")
        if 'status' in data:
            if data['status']=='running':
                task.status = 'running'
            elif data['status'] in ['pending','paused']:
                task.status = 'assigned'
            else:
                api.abort(500, f"A new execution cannot be created with a status {data['status']}")
        else:
            task.status='assigned'
        return super().create(data)

    def update(self, id, data):
        execution = self.get(id)
        modified = False
        for attr, value in data.items():
            if hasattr(execution,attr): 
                if getattr(execution,attr)!=value:
                    if attr=='status':
                        if value not in self.authorized_status:
                            api.abort(500,
                                f"Status {value} is not possible (only {' '.join(self.authorized_status)})")
                        task=task_dao.get(execution.task_id)
                        if execution.status=='pending':
                            if value=='running':
                                task.status = 'running'
                                task.modification_date = datetime.utcnow()
                            elif value=='accepted':
                                task.status = 'accepted'
                                task.modification_date = datetime.utcnow()
                            elif value in ['refused','failed']:
                                task.status = 'pending'
                                task.modification_date = datetime.utcnow()
                            elif value=='succeeded':
                                task.status = 'succeeded'
                                task.modification_date = datetime.utcnow()
                            else:
                                api.abort(500, f"An execution cannot change status from pending to {value}")
                        elif execution.status=='accepted':
                            if value in ['running','failed','succeeded']:
                                task.status = value
                                task.modification_date = datetime.utcnow()
                            else:
                                log.exception(f"An execution cannot change status from accepted to {value}")
                                api.abort(500, f"An execution cannot change status from accepted to {value}")
                        elif execution.status=='running':
                            if value in ['succeeded', 'failed']:
                                task.status=value
                                task.modification_date = datetime.utcnow()
                            else:
                                log.exception(f"An execution cannot change status from running to {value}")
                                api.abort(500, f"An execution cannot change status from running to {value}")
                        else:
                            api.abort(500, f"An execution cannot change status from {execution.status} (only from pending, running or accepted)")
                    setattr(execution, attr, value)
                    modified = True
            else:
                raise Exception('Error: {} has no attribute {}'.format(
                    execution.__name__, attr))
        if modified:
            execution.modification_date = datetime.utcnow()
            db.session.commit()
        return execution

    def list(self,*args):
        return super().list(*args,sorting_column='execution_id')

execution_dao = ExecutionDAO()

execution = api.model('Execution', {
    'execution_id': fields.Integer(readonly=True, description='The execution unique identifier'), 
    'worker_id': fields.Integer(required=True, description='A worker unique identifier'),
    'task_id': fields.Integer(required=True, description='A task unique identifier'),
    'status': fields.String(readonly=True,
        description=f'The execution status: {", ".join(ExecutionDAO.authorized_status)}'), 
    'return_code': fields.Integer(required=False, description='The return code of the execution (when finished)'),
    'pid': fields.Integer(required=False, description='The process id (pid) of the execution'),
    'creation_date': fields.DateTime(readonly=True,
        description="timestamp of execution creation (on worker)"),
    'modification_date': fields.DateTime(readonly=True,
        description="timestamp of execution last modification"),
    'output': fields.String(readonly=True, description='The standard output of the execution'),
    'error': fields.String(readonly=True, description='The standard error of the execution (if any)'),
    'output_files': fields.String(readonly=True, description='A list of output files transmitted (if any)'),
})


@ns.route("/<id>/executions")
@ns.param("id", "The worker identifier")
@ns.response(404, "Worker not found")
class WorkerExecutionObject(Resource):
    @ns.doc("get_worker_executions")
    @ns.marshal_list_with(execution)
    def get(self, id):
        """Fetch a worker executions given the worker identifier"""
        #worker_dao.update_contact(id)
        return execution_dao.list(Execution.worker_id==id)


@ns.route("/<id>/executions/<status>")
@ns.param("id", "The worker identifier")
@ns.param("status", "Filter only executions with this status")
@ns.response(404, "Worker not found")
class WorkerExecutionFilterObject(Resource):
    @ns.doc("get_worker_executions")
    @ns.marshal_list_with(execution)
    def get(self, id, status):
        """Fetch a worker executions given the worker identifier and the executions status"""
        #worker_dao.update_contact(id)
        return execution_dao.list(Execution.worker_id==id, Execution.status==status)

signal = api.model('Signal', {
    'execution_id': fields.Integer(readonly=True, description='The execution unique identifier'), 
    'worker_id': fields.Integer(readonly=True, description='A worker unique identifier'),
    'signal': fields.Integer(readonly=True, description='The signal to send to the execution (UNIX signal)'),
})

signal_parser = api.parser()
signal_parser.add_argument('execution_id', type=int, help='Execution identifier', location='json')
signal_parser.add_argument('signal', type=int, help='Signal to send', location='json')

@ns.route("/<id>/signals")
@ns.param("id", "The worker identifier")
@ns.response(404, "Worker not found")
class WorkerSignal(Resource):
    @ns.doc("get_worker_signals")
    @ns.marshal_list_with(signal)
    def get(self, id):
        """Fetch a worker executions given the worker identifier and the executions status"""
        signals = list(Signal.query.filter(Signal.worker_id==id))
        for sig in signals:
            db.session.delete(sig)
        db.session.commit()
        return signals
    
    @ns.doc("create_worker_signal")
    @ns.expect(signal_parser)
    def post(self, id):
        """Create a signal for the given worker about a certain execution with a certain signal"""
        args = signal_parser.parse_args()
        db.session.add(Signal(args['execution_id'], id, args['signal']))
        db.session.commit()


ns = api.namespace('executions', description='EXECUTION operations')

@ns.route('/')
class ExecutionList(Resource):
    '''Shows a list of all executions, and lets you POST to add new workers'''
    @ns.doc('list_executions')
    @ns.marshal_list_with(execution)
    def get(self):
        '''List all executions'''
        return execution_dao.list()

    @ns.doc('create_execution')
    @ns.expect(execution)
    @ns.marshal_with(execution, code=201)
    def post(self):
        '''Create a new execution'''
        return execution_dao.create(api.payload), 201
    

@ns.route("/<id>")
@ns.param("id", "The execution identifier")
@ns.response(404, "Execution not found")
class ExecutionObject(Resource):
    @ns.doc("get_execution")
    @ns.marshal_with(execution)
    def get(self, id):
        """Fetch a execution given its identifier"""
        return execution_dao.get(id)

    @ns.doc("update_execution")
    @ns.expect(execution)
    @ns.marshal_with(execution, code=201)
    def put(self, id):
        """Update an execution"""
        return execution_dao.update(id, api.payload)


parser = api.parser()
parser.add_argument('text', type=str, help='Supplementary text', location='json')

@ns.route("/<id>/delete")
@ns.param("id", "The execution identifier")
@ns.response(404, "Execution not found")
class ExecutionObject(Resource):
    @ns.doc('delete_execution')
    def put(self,id):
        '''Delete an execution for this id'''
        for e in Execution.query.filter(Execution.execution_id==id):
            db.session.delete(e)
        db.session.commit()
        return {'result':'ok'}

@ns.route("/<id>/output")
@ns.param("id", "The execution identifier")
@ns.response(404, "Execution not found")
class ExecutionOutput(Resource):
    @ns.doc("update_execution_output")
    @ns.expect(parser)
    def put(self, id):
        """Add some data to the execution output"""
        x = execution_dao.get(id)
        args = parser.parse_args()
        execution_dao.update(id, 
            {'output':('' if x.output is None else x.output) + args['text']})
        return {'result':'Ok'}

@ns.route("/<id>/error")
@ns.param("id", "The execution identifier")
@ns.response(404, "Execution not found")
class ExecutionOutput(Resource):
    @ns.doc("update_execution_error")
    @ns.expect(parser)
    def put(self, id):
        """Add some data to execution error"""
        x = execution_dao.get(id)
        args = parser.parse_args()
        execution_dao.update(id, 
            {'error':('' if x.error is None else x.error) + args['text']})
        return {'result':'Ok'}

@ns.route("/<id>/output_files")
@ns.param("id", "The execution identifier")
@ns.response(404, "Execution not found")
class ExecutionOutputFiles(Resource):
    @ns.doc("update_execution_output")
    @ns.expect(parser)
    def put(self, id):
        """Update output files (files transferred from /output, i.e. results) for execution as a space separated file URL list"""
        args = parser.parse_args()
        execution_dao.update(id, 
            {'output_files':args['text']})
        return {'result':'Ok'}


ns = api.namespace('batch', description='BATCH level operations')

batch_parser = api.parser()
batch_parser.add_argument('signal', type=int, help='Optionnaly send signal to all running tasks', 
    required=False, location='json')


custom_batch = api.model('Batch', {
    'batch': fields.String(readonly=True, description='The name of the batch'), 
    'pending': fields.Integer(readonly=True,
        description='The number of task in pending state'), 
    'accepted': fields.Integer(readonly=True,
        description='The number of task in accepted state'), 
    'running': fields.Integer(readonly=True,
        description='The number of task in running state'), 
    'failed': fields.Integer(readonly=True,
        description='The number of task in succeeded state'), 
    'succeeded': fields.Integer(readonly=True,
        description='The number of task in failed state'), 
    'workers': fields.String(readonly=True, description='The workers used in that batch'),
})

@ns.route("/")
class BatchList(Resource):
    @ns.doc("list_batch")
    @ns.marshal_list_with(custom_batch)
    def get(self):
        """List the batches, their task statuses and workers"""
        if isinstance(db.session.bind.dialect, sqlite.dialect):
            worker_query='''SELECT batch,GROUP_CONCAT(name,',') FROM worker GROUP BY batch'''
        else:
            worker_query='''SELECT batch,STRING_AGG(name,',') FROM worker GROUP BY batch'''
        batch_query='''SELECT batch,status,count(task_id) FROM task GROUP BY batch,status ORDER BY batch,status'''
        batches = []
        batches_attributes = {}
        for batch,status,count in db.session.execute(batch_query):
            if batch not in batches:
                batches.append(batch)
                batches_attributes[batch]={'batch':batch}    
            batches_attributes[batch][status]=count
        for batch,workers in db.session.execute(worker_query):
            batches_attributes[batch]['workers']=workers
        return [batches_attributes[batch] for batch in batches]
@ns.route("/<name>/stop")
@ns.param("name", "The batch name")
@ns.response(404, "Batch not found")
class BatchStop(Resource):
    @ns.doc("stop_a_batch")
    @ns.expect(batch_parser)
    def put(self, name):
        """Pause all workers for this batch - and kill current job if force is set"""
        if name=='Default':
            for w in Worker.query.filter(Worker.batch.is_(None)):
                w.status = 'paused'
        else:
            for w in Worker.query.filter(Worker.batch==name):
                w.status = 'paused'
        args = batch_parser.parse_args()
        if args.get('signal',False):
            log.warning(args['signal'])
            log.warning(f'Sending signal {args["signal"]} to executions for batch {name}')
            if name=='Default':
                for e in db.session.scalars(select(Execution).join(Execution.task).where(
                                                Execution.status=='running',
                                                Task.batch.is_(None))):
                    log.warning(f'Sending signal {args["signal"]} to execution {e.execution_id}')
                    db.session.add(Signal(e.execution_id, e.worker_id, args['signal']))
            else:
                for e in db.session.scalars(select(Execution).join(Execution.task).where(
                                                Execution.status=='running',
                                                Task.batch==name)):
                    log.warning(f'Sending signal {args["signal"]} to execution {e.execution_id}')
                    db.session.add(Signal(e.execution_id, e.worker_id, args['signal']))
        db.session.commit()
        return {'result':'Ok'}

@ns.route("/<name>/go")
@ns.param("name", "The batch name")
@ns.response(404, "Batch not found")
class BatchGo(Resource):
    @ns.doc("_re_launch_a_batch")
    @ns.expect(batch_parser)
    def put(self, name):
        """(re)set all workers affected to this batch to running"""
        if name=='Default':
            for w in Worker.query.filter(Worker.batch.is_(None)):
                w.status = 'running'
        else:
            for w in Worker.query.filter(Worker.batch==name):
                w.status = 'running'
        args = batch_parser.parse_args()
        if args.get('signal',False):
            log.warning(args['signal'])
            log.warning(f'Sending signal {args["signal"]} to executions for batch {name}')
            if name=='Default':
                for e in db.session.scalars(select(Execution).join(Execution.task).where(
                                                Execution.status=='running',
                                                Task.batch.is_(None))):
                    log.warning(f'Sending signal {args["signal"]} to execution {e.execution_id}')
                    db.session.add(Signal(e.execution_id, e.worker_id, args['signal']))
            else:
                for e in db.session.scalars(select(Execution).join(Execution.task).where(
                                                Execution.status=='running',
                                                Task.batch==name)):
                    log.warning(f'Sending signal {args["signal"]} to execution {e.execution_id}')
                    db.session.add(Signal(e.execution_id, e.worker_id, args['signal']))
        db.session.commit()
        return {'result':'Ok'}

@ns.route("/<name>")
@ns.param("name", "The batch name")
@ns.response(404, "Batch not found")
class BatchDelete(Resource):
    @ns.doc("delete_a_batch")
    def delete(self, name):
        """Delete all tasks and executions for this batch"""
        # execution are deleted by cascade
        if name=='Default':
            for t in Task.query.filter(Task.batch.is_(None)):
                db.session.delete(t)
        else:
            for t in Task.query.filter(Task.batch==name):
                db.session.delete(t)
        db.session.commit()
        return {'result':'Ok'}

#     # ### 
#     #  #  
#     #  #  
#     #  #  
#     #  #  
#     #  #  
 #####  ### 



@app.route('/ui/')
def ui(name=None):
    return render_template('ui.html', name=name)
@app.route('/ui/task/')
def task():
    return render_template('task.html')

@app.route('/ui/batch/')
def batch():
    return render_template('batch.html')

#@app.route('/test/')


@socketio.on('get')
def handle_get(json):
    if 'delay' in json:
        sleep(int(json['delay']))
    if json['object']=='workers':
        log.info('sending workers')
        #emit('workers', [(w.name, w.status, len(w.executions)) for w in Worker.query.all()])
        emit('workers', {
            'workers':list([dict(row) for row in db.session.execute(
                '''SELECT 
                    worker_id,
                    name, 
                    batch,
                    status, 
                    concurrency,
                    prefetch,
                    (SELECT count(execution_id) FROM execution WHERE execution.worker_id=worker.worker_id AND
                                                                        execution.status='accepted') as accepted,
                    (SELECT count(execution_id) FROM execution WHERE execution.worker_id=worker.worker_id AND 
                                                                        execution.status='running') as running,
                    (SELECT count(execution_id) FROM execution WHERE execution.worker_id=worker.worker_id AND 
                                                                        execution.status='succeeded') as succeeded,
                    (SELECT count(execution_id) FROM execution WHERE execution.worker_id=worker.worker_id AND 
                                                                        execution.status='failed') as failed,
                    (SELECT count(execution_id) FROM execution WHERE execution.worker_id=worker.worker_id) as total,
                    load,
                    memory,
                    read_bytes,
                    written_bytes 
                FROM worker
                ORDER BY worker.batch,worker.name''')]),
            'totals': list([list(row) for row in db.session.execute(
                '''SELECT
                    (SELECT count(task_id) FROM task WHERE status='pending'),
                    (SELECT count(task_id) FROM task WHERE status IN ('assigned','accepted')),
                    (SELECT count(task_id) FROM task WHERE status='running'),
                    (SELECT count(task_id) FROM task WHERE status='failed'),
                    (SELECT count(task_id) FROM task WHERE status='succeeded')
                '''
            )])[0] })

    if json['object']=='task':
        log.warning(f"sending task ordered by {json['order_by']} filter by {json['filter_by']}")

        if json['order_by'] =='worker':
            sort_clause='ORDER BY worker.name, task.task_id DESC'
        elif json['order_by'] =='batch':
            sort_clause='ORDER BY task.batch, task.task_id DESC'
        else:
            sort_clause='ORDER BY task.task_id DESC'
        
        if json['filter_by']=='terminated':
            where_clause = "WHERE execution.status IN ('succeeded','failed')"
        elif json['filter_by']=='all':
            where_clause = ""
        elif json['filter_by']:
            where_clause = f"WHERE execution.status='{json['filter_by']}'"
        else:
            where_clause = ""

        if IS_SQLITE:
            trunc_output=f'SUBSTR(execution.output,-{UI_OUTPUT_TRUNC},{UI_OUTPUT_TRUNC})'
            trunc_error=f'SUBSTR(execution.error,-{UI_OUTPUT_TRUNC},{UI_OUTPUT_TRUNC})'
        else:
            trunc_output=f'RIGHT(execution.output,{UI_OUTPUT_TRUNC})'
            trunc_error=f'RIGHT(execution.error,{UI_OUTPUT_TRUNC})'
        
        task_list = list([list(map(lambda x : str(x) if type(x)==type(datetime.utcnow()) else x ,row)) for row in db.session.execute(
        f'''SELECT
        task.task_id,
        task.name,
        worker.name,
        task.batch,
        execution.creation_date,
        execution.modification_date,
        execution.execution_id,
        {trunc_output},
        {trunc_error},
        task.command,
        execution.worker_id,
        task.status
        FROM task 
        LEFT JOIN execution ON (task.task_id=execution.task_id 
            AND execution.creation_date=(SELECT MAX(creation_date) FROM execution AS e1 WHERE e1.task_id=task.task_id))
        LEFT JOIN worker ON execution.worker_id=worker.worker_id 
        {where_clause}
        {sort_clause}
        '''
        )])

        if json.get('detailed_tasks',None):
            for detailed_task in db.session.execute(f"""
                SELECT execution_id,output,error FROM execution 
                WHERE execution_id IN ({','.join([str(eid) for eid in json['detailed_tasks']])})"""):
                for task in task_list:
                    if task[6]==detailed_task[0]:
                        task[7]=detailed_task[1]
                        task[8]=detailed_task[2]
                        break

        emit('task',{'tasks':task_list})
        
    #if json['object'] in ['succeeded','failed','running','pending']:
    #    log.info('sending executions')
    #    type_execution = json['object']
    #    emit('execution',{'executions':list([list(row) for row in db.session.execute(
    #    f'''SELECT 
    #    execution_id,
    #    worker_id,
    #    task_id,
    #    status,
    #    return_code,
    #    pid,
    #    output,
    #    error
    #    FROM execution
    #    WHERE execution.status 
    #    LIKE '{type_execution}'

    #   '''
    #    )])})
    if json['object'] == 'batch':
        log.info('sending batch')
        if IS_SQLITE:
            duration_query='(JULIANDAY(e1.modification_date)-JULIANDAY(e1.creation_date))*24'
            worker_query='''SELECT batch,GROUP_CONCAT(name,',') FROM worker GROUP BY batch'''
        else:
            duration_query='EXTRACT ( EPOCH FROM (e1.modification_date-e1.creation_date)/3600 )'
            worker_query='''SELECT batch,STRING_AGG(name,',') FROM worker GROUP BY batch'''
        batch_query=f'''SELECT * FROM (
    SELECT batch,status,COUNT(task_id),MAX(duration),MIN(duration), AVG(duration) FROM (
        SELECT {duration_query} as duration, e1.task_id, e1.status,task.batch FROM execution e1 JOIN task ON (
            task.task_id=e1.task_id AND e1.creation_date=(
                SELECT MAX(creation_date) FROM execution WHERE execution.task_id=task.task_id
            )
        )
    ) AS e2 GROUP BY batch,status
    UNION	 
    SELECT batch,status, COUNT(task_id),NULL,NULL,NULL 
    FROM task WHERE task_id NOT IN (SELECT task_id FROM execution) GROUP BY batch,status
) AS b ORDER BY batch, status'''
        emit('batch',{'batches':list([list(map(lambda x : str(x) if type(x)==type(datetime.utcnow()) else x ,row)) for row in db.session.execute(
        batch_query
        )]),
        'workers': list([list(row) for row in db.session.execute(worker_query)])})
        
@socketio.on('change_batch')
def handle_change_batch(json):
    Worker.query.filter(Worker.worker_id==json['worker_id']).update(
        {Worker.batch:json['batch_name'] or None})
    db.session.commit()


@socketio.on('concurrency_change')
def handle_concurrency_change(json):
    worker_id = json['id']
    change = json['change']
    log.info(f'changing concurrency for worker {worker_id}: {change}')
    if isinstance(db.session.bind.dialect, sqlite.dialect):
        log.info('Using sqlite SQL')
        Worker.query.filter(Worker.worker_id==worker_id).update(
            {Worker.concurrency: func.max(Worker.concurrency+change,0)})
    else:
        log.info('Using standard SQL')
        Worker.query.filter(Worker.worker_id==worker_id).update(
            {Worker.concurrency: func.greatest(Worker.concurrency+change,0)})
    db.session.commit()

@socketio.on('prefetch_change')
def handle_prefetch_change(json):
    worker_id = json['id']
    change = json['change']
    log.info(f'changing prefetch for worker {worker_id}: {change}')
    if isinstance(db.session.bind.dialect, sqlite.dialect):
        log.info('Using sqlite SQL')
        Worker.query.filter(Worker.worker_id==worker_id).update(
            {Worker.prefetch: func.max(Worker.prefetch+change,0)})
    else:
        log.info('Using standard SQL')
        Worker.query.filter(Worker.worker_id==worker_id).update(
            {Worker.prefetch: func.greatest(Worker.prefetch+change,0)})
    db.session.commit()

def get_nodename():
    worker_names = list(map(lambda x: x[0], 
        db.session.execute(select(Worker.name))))
    log.warning(f'Worker names: {worker_names}')
    i=1
    while f'node{i}' in worker_names:
        i+=1
    return f'node{i}'



@socketio.on('create_worker')
def handle_create_worker(json):
    concurrency = int(json['concurrency'])
    flavor = json['flavor']
    if not flavor:
        emit('worker_created', 'Flavor must be specified')
        return None
    region = json['region']
    if not region:
        emit('worker_created', 'Region must be specified')
        return None
    batch = json['batch'] or None
    prefetch = int(json['prefetch'])
    number = int(json['number'])
    global worker_create_queue
    for _ in range(number):
        worker_create_queue.put(Namespace(concurrency=concurrency, 
            prefetch=prefetch,
            flavor=flavor,
            region=region,
            batch=batch))

@socketio.on('batch_action')
def handle_batch_action(json):
    """Gathering all the action dealing with batch like pause, break, stop, clear, go."""
    if json['action'] in ['stop','break','pause','simple pause','pause only batch']:
        #Same function as in the API set all workers affected to this batch to running and can also interrupt the running tasks with signal 3 and 9
        name=json['name']
        if name=='Default':
            name=None
        for w in Worker.query.filter(Worker.batch==name):
                w.status = 'paused'
        if json['action']=='break':
            signal = 9
        elif json['action']=='stop':
            signal = 3
        elif json['action']=='simple pause':
            signal = 0
        elif json['action']=='pause':
            for t in Task.query.filter(and_(
                        Task.batch==name,
                        Task.status.in_(['running','accepted']))):
                t.status = 'paused'
            signal = 20
            db.session.commit()
        log.warning(f'Sending signal {signal} to executions for batch {name}')
        for e in db.session.scalars(select(Execution).join(Execution.task).where(
                                        Execution.status=='running',
                                        Task.batch==name)):
            log.warning(f'Sending signal {signal} to execution {e.execution_id}')
            db.session.add(Signal(e.execution_id, e.worker_id, signal ))
        db.session.commit()
        log.warning('result pause :Ok')
    if json['action'] in ['go','simple go']: 
        #Same function as in the API (re)set all workers affected to this batch to running
        name=json['name']
        """(re)set all workers affected to this batch to running"""
        for w in Worker.query.filter(Worker.batch==name):
            w.status = 'running'
        if json['action']=='go':
            signal=18
            for t in Task.query.filter(Task.batch==name):
                if t.status == 'paused':
                    t.status='running'
            for e in db.session.scalars(select(Execution).join(Execution.task).where(
                                            Execution.status=='running',
                                            Task.batch==name)):
                log.warning(f'Sending signal {signal} to execution {e.execution_id}')
                db.session.add(Signal(e.execution_id, e.worker_id, signal ))
        db.session.commit()
        log.warning('result go : Ok')
    if json['action']=='clear':
        #Same function as in the API clear() Delete all tasks and executions for this batch
        name=json['name']
        """Delete all tasks and executions for this batch"""
        # execution are deleted by cascade
        if name=='Default':
            name=None
        for t in Task.query.filter(Task.batch==name):
            db.session.delete(t)
        db.session.commit()
        log.warning(f'result clear batch {name}: Ok ')

@socketio.on('task_action')
def handle_task_action(json):
    """Gathering all the action dealing with task like break, stop, delete, modify, restart"""
    #The code essentially is from the API code with a few modifications
    task=json['task_id']
    if json['action'] in ['break','stop','pause','resume']: 
        #A signal 3 or 9 is created and causes only the interruption of the task with id same structure as in the API 
        task=json['task_id']
        for t in Task.query.filter(Task.task_id==task):
            if json['action'] == 'break':
                type='break'
                signal = 9
            elif json['action'] == 'stop':
                signal = 3
                type='stop'
            elif json['action'] == 'pause':
                signal = 20
                type='pause'
                t.status='paused'
            elif json['action'] == 'resume':
                signal = 18
                type='resume'
                t.status='running'
            log.warning(f'Sending signal {signal} to executions for task {task}')
            for e in db.session.scalars(select(Execution).join(Execution.task).where(
                                            Execution.status=='running',
                                            Task.task_id==task)):
                log.warning(f'Sending signal {signal} to execution {e.execution_id}')
                db.session.add(Signal(e.execution_id, e.worker_id, signal ))
        db.session.commit()
        log.warning(f'result {type} : Ok')
    if json['action']=='delete': 
        #Delete the task in the data base
        for t in Task.query.filter(Task.task_id==task):
            db.session.delete(t)
        db.session.commit()
        log.warning('result delete: Ok')
    if json['action']=='modify': 
        #Changing the command for a task in the data base and moving it in the task queue. It doesn't create a new task.
        for t in Task.query.filter(Task.task_id==task):
            t.command =json["modification"]
            t.status='pending'
        db.session.commit()
        log.warning('result modify : Ok')
    if json['action']=='restart': 
        #Relaunching the execution of a task.
        for t in Task.query.filter(Task.task_id==task):
            t.status='pending'
        db.session.commit()
        log.warning('result restart : Ok')

@socketio.on('delete_worker') #Delete a worker.
def delete_worker(json):
    """Delete a worker in db"""
    return worker_dao.delete(json['worker_id'])
    

def create_worker_object(concurrency, flavor, region, batch, prefetch, db_session):
    """Create a worker object in db - this must be called linearly not in async way
    """
    hostname = get_nodename()
    idle_callback = WORKER_IDLE_CALLBACK.format(hostname=hostname)
    log.info(f'Creating a new worker {hostname}: concurrency:{concurrency}, flavor:{flavor}, region:{region}, prefetch:{prefetch}')
    w = Worker(name=hostname, hostname=hostname, concurrency=concurrency, status='offline', 
            batch=batch, idle_callback=idle_callback, prefetch=prefetch)
    db_session.add(w)
    db_session.commit()
    return hostname, w
    
    
def create_worker_process(hostname, concurrency, flavor, region, batch, w, 
        db_session, prefetch=None):
    """Deploy a worker - can be called parallely up to a certain limit
    (WORKER_CREATE_CONCURRENCY)
    """
    try:
        regions = region.split()
        if len(regions)>1:
            region=regions[0]
            len(f'Splitting regions and using region {region}')
        retry=WORKER_CREATE_RETRY
        # be careful with lazy objects in case they got deleted
        worker_id,worker_name = w.worker_id,w.name
        while retry>=0 and worker_exists(worker_id, db_session):
            log.warning(f'''Command is {WORKER_CREATE.format(
                hostname=hostname,
                concurrency=concurrency,
                flavor=flavor,
                region=region
            )}''')
            process = run([WORKER_CREATE.format(
                hostname=hostname,
                concurrency=concurrency,
                flavor=flavor,
                region=region
            )],shell=True, capture_output=True, encoding='utf-8')
            if process.returncode == 0:
                log.warning(f'worker created: done for {hostname}')
                socketio.emit('worker_created', f'done for {hostname}',broadcast=True)
                db_session.close()
                break
            else:
                log.warning(f'worker not created: error for {hostname}: {process.stdout.strip()} {process.stderr.strip()}')
                socketio.emit('worker_created', f'error for {hostname}: {process.stdout.strip()} {process.stderr.strip()}',broadcast=True)
                log.warning(f'Broadcast message sent for {worker_name} failure')
                if retry>=0:
                    log.warning('Retrying...')
                    sleep(WORKER_CREATE_RETRY_SLEEP)
                else:
                    db_session.close()
            #if w.idle_callback:
            #    log.warning(f'Worker {w.name} called idle callback, launching: '+w.idle_callback.format(**(worker.__dict__)))
            #    process = run([w.idle_callback.format(**(w.__dict__))],shell=True,capture_output=True, encoding='utf-8')
            #    if process.returncode != 0:
            #        log.warning(f'worker callback failed: error for {hostname}: {process.stdout.strip()} {process.stderr.strip()}')
            #    else:
            #        log.warning(f'Worker {w.name} was successfully undeployed')
            #        if len(regions)>1:
            #            log.warning("Trying next regions")
            #            create_worker_process(hostname, concurrency, flavor, " ".join(regions[1:]), batch, w)
    except Exception as e:
        socketio.emit('worker_created', f'error for {hostname}: {e}',broadcast=True)
        log.exception(f'Creation failed for {hostname}...')
        raise

def worker_exists(worker_id, db_session):
    """Return true if this worker_id is in database"""
    return db_session.query(Worker).filter(Worker.worker_id==worker_id).count()>0

#@app.route('/ui/get/')
#def ui_get():
#    object = request.args.get('object','workers')
#    log.info('Return worker list to UI')
#    if object=='workers':
#        log.info('sending workers')
#        return jsonify(workers = list([list(row) for row in db.session.execute(
#            'SELECT name, status, (SELECT count(execution_id) FROM execution WHERE execution.worker_id=worker.worker_id) FROM worker')]))

def background():
    # while some tasks are pending without executions:
    #   look for a running worker:
    #      create a pending execution of this task for this worker
    with app.app_context():
        session = Session(db.engine)
    worker_create_process_waiting_queue = []
    worker_create_process_queue = []
    log.info('Starting thread for {}'.format(os.getpid()))
    while True:
        log.info('Starting main loop')
        task_attributions = False
        try:
            batch_already_dispatched = []
            task_list = list(session.query(Task).filter(
                    Task.status=='pending').with_entities(Task.task_id, Task.batch))
            if task_list:
                worker_list = list(session.query(Worker).filter(
                            Worker.status=='running').with_entities(
                            Worker.worker_id,Worker.batch,Worker.concurrency,Worker.prefetch))
                execution_per_worker = {worker_id: count for worker_id,count in 
                                session.query(Execution.worker_id,func.count(Execution.task_id)).filter(and_(
                                    Execution.worker_id.in_(list([w.worker_id for w in worker_list])),
                                    Execution.status.in_(['running','pending','accepted']))
                                ).group_by(
                                    Execution.worker_id
                                )
                }
                
                for task in task_list:
                    if task.batch in batch_already_dispatched:
                        continue
                    for worker in worker_list:
                        if worker.batch != task.batch:
                            continue
                        if execution_per_worker.get(worker.worker_id,0)<(worker.concurrency+worker.prefetch):
                            session.add(Execution(worker_id=worker.worker_id,
                                task_id=task.task_id))
                            session.query(Task).filter(Task.task_id==task.task_id).update(
                                {'status':'assigned'}
                            )
                            execution_per_worker[worker.worker_id] = execution_per_worker.get(worker.worker_id,0)+1
                            log.info(f'Execution of task {task.task_id} proposed to worker {worker.worker_id}')
                            task_attributions = True
                            break
                    else:
                        log.info(f'All worker are full for batch {task.batch}, giving up for this round')
                        batch_already_dispatched.append(task.batch)
            if task_attributions:
                session.commit()
            now = datetime.utcnow()
            log.warning('Looking for offline/online workers')
            session.expire_all()
            for worker in session.query(Worker).filter(Worker.status.in_(['offline','running'])):
                change = False
                log.warning(f'Assessing worker {worker.name}')
                if worker.last_contact_date is None or (
                            now - worker.last_contact_date
                        ).total_seconds() > WORKER_OFFLINE_DELAY:
                    log.warning(f'... late: {worker.last_contact_date}')
                    if worker.status=='running':
                        log.warning(f'Worker {worker.name} ({worker.worker_id}) lost, marked as offline')
                        worker.status = 'offline'
                        change = True
                else:
                    log.warning('... good')
                    if worker.status=='offline':
                        log.warning(f'Worker {worker.name} ({worker.worker_id}) recovered, marked as running')
                        worker.status = 'running'
                        change = True
                if change:
                    session.commit()  
            try:
                while True:
                    new_worker = worker_create_queue.get(block=False)
                    new_worker.hostname, new_worker.w = create_worker_object(
                        concurrency=new_worker.concurrency,
                        flavor=new_worker.flavor,
                        region=new_worker.region,
                        batch=new_worker.batch,
                        prefetch=new_worker.prefetch,
                        db_session=session)
                    worker_create_process_waiting_queue.append(
                        new_worker
                    )
            except queue.Empty:
                pass
            while len(worker_create_process_waiting_queue)>0 \
                    and len(worker_create_process_queue)<WORKER_CREATE_CONCURRENCY:
                new_worker = worker_create_process_waiting_queue.pop(0)
                with app.app_context():
                    new_worker.db_session = Session(db.engine)
                # verifying that the object is still in db to prevent recreation
                if worker_exists(new_worker.w.worker_id, session):
                    log.warning(f'Launching creation process for worker {new_worker.hostname}.')
                    worker_create_process = PropagatingThread(
                        target = create_worker_process,
                        kwargs = new_worker.__dict__)
                    worker_create_process.start()
                    worker_create_process_queue.append((new_worker, worker_create_process))
                else:
                    log.warning(f'Not queuing {new_worker.hostname} for creation as it was deleted')
            for new_worker,worker_create_process in list(worker_create_process_queue):
                try:
                    if not worker_create_process.is_alive():
                        worker_create_process_queue.remove((new_worker, worker_create_process))
                        log.warning(f'Creation process over for worker {new_worker.hostname}.')
                except Exception as e:
                    log.exception(f'Creation failed: {e}')
                    worker_create_process_queue.remove((new_worker, worker_create_process))
                    

            



                
        except Exception:
            log.exception('An exception occured during server main loop:')
            while True:
                sleep(MAIN_THREAD_SLEEP)
                try:
                    log.warning('Trying to reconnect...')
                    with app.app_context():
                        session = Session(db.engine)
                    break
                except Exception:
                    pass
        sleep(MAIN_THREAD_SLEEP)


Thread(target=background).start()


def main():
    #app.run(debug=False)
    socketio.run(app)

if __name__ == "__main__":
    raise RuntimeError('Do not launch directly, launch with "FLASK=scitq.server flask run"')


import re
import requests
from requests.exceptions import Timeout, ConnectionError
from datetime import datetime
from time import sleep
import threading
import queue
from argparse import Namespace
import logging as log


PUT_TIMEOUT = 30
GET_TIMEOUT = 150
QUERY_THREAD_TIMEOUT = 10
JOIN_DYNAMIC_SLEEP_TIME_INCREMENT = 10

def _parse_date_andco(item):
    """A custom filter to transform JSON date (i.e. date in ISO formated text) 
    in datetime object. JSON object can be list of dicts or a simple dict or
    a simple object directly in some case. Only variable with name ending with 
    _date are converted (thus only within a dictionary, standalone object are
    not converted)
    Perform also misc transformation like batch default"""
    def _sub_parse(subitem):
        for key, value in subitem.items():
            if '_date' in key and value is not None:
                subitem[key] = datetime.fromisoformat(value)
            elif key=='batch' and value is None:
                subitem[key] = 'Default'
        return subitem
    if type(item)==list:
        return list([_sub_parse(subitem) for subitem in item])
    elif type(item)==dict:
        return _sub_parse(item)
    else:
        return item

def _clean(d):
    """filter out null value from a dict as well as key in remove list (default
    to ['id']) plus a ugly hack for default batch"""
    return dict([('batch',None) if (k=='batch' and v=='Default') else (k,v)  
        for k,v in d.items() if v is not None])

def _to_obj(d):
    """A simple recursive wrapper to convert JSON like objects in python objects"""
    if type(d)==list:
        return map(_to_obj, d)
    if type(d)==dict:
        return Namespace(**d)
    return d

class HTTPException(Exception):
    def __init__(self, text, message, status_code):
        self.status_code = status_code
        self.message = message
        super().__init__(text)


def _filter_non200(r):
    if r.status_code//100 != 2:
        try:
            message = r.json()['message']
        except:
            message = 'Unknown server error'
        raise HTTPException(f'Error {r.status_code}: {message}', 
                message=message, 
                status_code=r.status_code)
    return r.json()
class LazyObject:
    """A lazy object or dict initialized with a queue.Queue 
    perform q.get as soon as you try to use it..."""
    def __init__(self, q):
        """Initialize object with queue.Queue q"""
        self._q = q
        self._o = None
    def __getattr__(self, __name):
        """Retrieve real attribute (object style) in real object once it is there"""
        if self._o is None:
            self._o = self._q.get()
        return getattr(self._o, __name)
    def __getitem__(self, __key):
        """Retrieve real key in real object (dict style) once it is there"""
        if self._o is None:
            self._o = self._q.get()
        return self._o[__key]
    def __str__(self):
        """Wait for real object and use real object representation"""
        if self._o is None:
            self._o = self._q.get()
        return str(self._o)


def query_thread(semaphore,send_queue, put_timeout):
        """A Query-thread which treats all the queries with a while loop until the object is there, then it put the result in the returning queue
        """
        while True :
            while not semaphore.acquire(blocking=False):
                try:
                    self,type,query,return_queue = send_queue.get(timeout=QUERY_THREAD_TIMEOUT)
                    break
                except queue.Empty:
                    continue
            else: 
                #This block is executed when the semaphore is acquired (meaning the Server instance was closed)
                break
            url,data=query
            task_done=False
            if type == 'put':
                while not task_done:
                    try:
                        sleep(put_timeout) 
                        result = self._wrap(requests.put(
                                self.url+url, json=data, timeout=put_timeout))
                        task_done = True
                    except (ConnectionError,Timeout) as e:
                        log.warning(f'Exception when trying to put: {e}')
                return_queue.put(result)
            elif type == 'post':
                while not task_done:
                    try:
                        sleep(put_timeout)
                        result = self._wrap(requests.post(
                                self.url+url, json=data, timeout=put_timeout))
                        task_done = True
                    except (ConnectionError,Timeout) as e:
                        log.warning(f'Exception when trying to post: {e}') 
                return_queue.put(result)
            elif type == 'delete':
                while not task_done:
                    try:
                        sleep(put_timeout)
                        result = self._wrap(requests.delete(
                                self.url+url, timeout=put_timeout))
                        task_done = True
                    except (ConnectionError,Timeout) as e:
                        log.warning(f'Exception when trying to post: {e}') 
                return_queue.put(result)            

class Server:
    """A thin wrapper above requests with URLs registered
    and several extrathings:
    - output is JSON de-marshalled,
    - date are converted,
    - if server is not accessible, the lib will retry or do the job later,
    - expected arguments are explicit which makes life easier.
    
    Arguments:
    - ip: the name or IP address of server,
    - style: if 'dict' (default) all the objects are rendered as dict, if 
        'object', dict are passed to Namespace to produce real Python objects.
    """

    def __init__(self, ip, style='dict', asynchronous=True, 
            put_timeout=PUT_TIMEOUT, get_timeout=GET_TIMEOUT):
        """initialise the object with IP or name of the server
        
        Arguments:
        - ip: the name or IP address of server,
        - style: if 'dict' (default) all the objects are rendered as dict, if 
            'object', dict are passed to Namespace to produce real Python objects.
        - asynchronous: if True (default) then put and post operations are 
            asynchronous, if False then all operations will wait (forever) for 
            the server to come back. get operations are always synchronous"""
        self.ip=ip
        self.url=f'http://{ip}:5000'
        if style=='dict':
            self._wrap = lambda x: _parse_date_andco(_filter_non200(x))
        elif style=='object':
            self._wrap = lambda x: _to_obj(_parse_date_andco(_filter_non200(x)))
        else:
            raise Exception(f'style can only be dict or object not f{style}')
        self.asynchronous = asynchronous
        if self.asynchronous:
            self.query_thread_semaphore = threading.Semaphore()
            self.query_thread_semaphore.acquire(blocking=False)
            self.send_queue= queue.Queue()
            self.query_thread=threading.Thread(target=query_thread,
                args=(self.query_thread_semaphore,self.send_queue,put_timeout))
        self.get_timeout = get_timeout
        self.put_timeout = put_timeout

    def queue_size(self):
        """A method to estimate send queue size"""
        return self.send_queue.qsize()        

    def get(self, url):
        """A wrapper used for all get operations.
        - url: extra string to add after base server URL
        
        return the objects according to Server style (see style in class doc)"""
        try:
            return self._wrap(requests.get(
                self.url+url, timeout=self.get_timeout
            ))
        except (ConnectionError,Timeout) as e:
            log.warning(f'Exception when trying to get: {e}')
            sleep(self.get_timeout)
            return self.get(url)

    def put(self,url, data):
        """A wrapper used for all put operations.
        - url: extra string to add after base server URL
        - data: extra data (payload of put operation) represented as dict

        return the objects according to Server style (see style in class doc)
        (in asynchronous mode, return a lazy object)"""
        try:
            return self._wrap(requests.put(
                self.url+url, json=data, timeout=self.put_timeout
            ))
        except (ConnectionError,Timeout) as e:
            if not self.query_thread.is_alive():
                self.query_thread.start()
            log.warning(f'Exception when trying to put: {e}')
            sleep(self.put_timeout)
            return_queue=queue.Queue()
            self.send_queue.put((self,'put',(url, data),return_queue))
            return LazyObject(return_queue)
                
        

    def post(self, url, data):
        """A wrapper used for all post operations.
        - url: extra string to add after base server URL
        - data: extra data (payload of post operation) represented as dict

        return the objects according to Server style (see style in class doc)"""
        try:
            return self._wrap(requests.post(
                url=self.url+url, json=data, timeout=self.put_timeout
            ))
        except (ConnectionError,Timeout) as e:
            if not self.query_thread.is_alive():
                self.query_thread.start()
            log.warning(f'Exception when trying to post: {e}')
            return_queue=queue.Queue()
            self.send_queue.put((self,'post',(url, data),return_queue))
            return LazyObject(return_queue)

    def delete(self,url):
        """A wrapper used for all put operations.
        - url: extra string to add after base server URL
        - data: extra data (payload of put operation) represented as dict

        return the objects according to Server style (see style in class doc)
        (in asynchronous mode, return a lazy object)"""
        try:
            return self._wrap(requests.delete(
                self.url+url, timeout=self.put_timeout
            ))
        except (ConnectionError,Timeout) as e:
            if not self.query_thread.is_alive():
                self.query_thread.start()
            log.warning(f'Exception when trying to put: {e}')
            sleep(self.put_timeout)
            return_queue=queue.Queue()
            self.send_queue.put((self,'delete',(url, None),return_queue))
            return LazyObject(return_queue)


    def workers(self):
        """Get a list of all workers"""
        return self.get('/workers/')

    def worker_update(self, id, name=None, hostname=None, concurrency=None, 
                    status=None,batch=None, idle_callback=None, prefetch=None):
        """Update a specific worker with worker_id equal to id
        return the updated worker (or None if the server timeout)"""
        return self.put(f'/workers/{id}', data=_clean(
            {'name':name, 'hostname':hostname, 'concurrency':concurrency, 
            'status':status, 'batch':batch, 'idle_callback':idle_callback,
            'prefetch':prefetch}
        ))

    def worker_create(self, name, concurrency, hostname=None, status='paused',
                batch=None, idle_callback=None, prefetch=0):
        """Create a new worker
        return the new worker (or None if the server timeout)"""
        return self.post('/workers/', data=_clean(
            {'name':name, 'hostname':hostname, 'concurrency':concurrency, 
            'status':status, 'batch':batch, 'idle_callback':idle_callback,
            'prefetch':prefetch}
        ))

    def worker_get(self, id):
        """get a specific worker with worker_id equal to id
        return the worker"""
        return self.get(f'/workers/{id}')

    def worker_ping(self, id, load, memory, read_bytes, written_bytes):
        """Update a specific worker ping time (heartbit)
        return a updated worker object with attribute or key (depending on style)"""
        return self.put(f'/workers/{id}/ping', data={'load':load,'memory':memory,'read_bytes':read_bytes,'written_bytes':written_bytes})

    def worker_callback(self, id, message):
        """Send a callback message (mainly idle) to trigger action on worker from the server
        return a object with result attribute, equal to ok"""
        return self.put(f'/workers/{id}/callback', data={'message':message})
    
    def worker_executions(self, id, status=None):
        """Get a list of a worker current assigned executions. 
        If status is used, only executions with this status are shown
        """
        if status is None:
            return self.get(f'/workers/{id}/executions')
        else:
            return self.get(f'/workers/{id}/executions/{status}')

    def worker_delete(self, id):
        """Delete a worker"""
        return self.delete(f'/workers/{id}')

    def workers_tasks(self):
        """Return a list of all the different tasks for a certain worker"""
        return self.get(f'/workers/tasks')
    
    def executions(self):
        """Get a list of all executions. 
        """
        return self.get(f'/executions/')
    
    def execution_create(self, worker_id, task_id, status='pending'):
        """Create a new execution, return the newly created execution
        """
        return self.post('/executions/', data={
            'worker_id':worker_id, 'task_id':task_id, 'status':status
        })

    def execution_update(self, id, status=None, pid=None, return_code=None, 
                        output=None, output_files=None):
        """Update a specific execution, return the updated execution
        """
        return self.put(f'/executions/{id}', data=_clean(
            {'status':status, 'pid':pid, 'return_code':return_code, 
                'output':output, 'output_files':output_files}
        ))

    def execution_output_write(self, id, output):
        """Add some output to a specific execution, return the updated execution
        """
        if output is None:
            return None
        return self.put(f'/executions/{id}/output', data=_clean(
            {'text':output}
        ))

    def execution_error_write(self, id, error):
        """Add some error (stderr) to a specific execution, return the updated execution
        """
        if error is None:
            return None
        return self.put(f'/executions/{id}/error', data=_clean(
            {'text':error}
        ))

    def execution_get(self, id):
        """get a specific execution with execution_id equal to id
        return the execution"""
        return self.get(f'/executions/{id}')

    def task_create(self, command, name=None, status='pending',batch=None, 
            input=None, output=None, container=None, container_options='',
            resource=None):
        """Create a new task, return the newly created task
        """
        return self.post('/tasks/', data=_clean({
            'command':command, 'name':name, 'status':status, 'batch':batch,
            'input':input, 'output':output, 'container':container, 
            'container_options':container_options, 'resource':resource
        }))

    def task_update(self, id, command=None, name=None, status=None, batch=None, 
            input=None, output=None, container=None, container_options=None,
            resource=None):
        """Update a specific execution, return the updated execution
        """
        return self.put(f'/tasks/{id}', data=_clean({
            'command':command, 'name':name, 'status':status, 'batch':batch,
            'input':input, 'output':output, 'container':container, 
            'container_options':container_options, 'resource':resource
        }))

    def task_get(self, id):
        """get a specific task with task_id equal to id
        return the task"""
        return self.get(f'/tasks/{id}')

    def tasks(self):
        """Get a list of all tasks. 
        """
        return self.get(f'/tasks/')

    def task_delete(self,id):
        """delete a specific task"""
        return self.delete(f'/tasks/{id}')

    def worker_signals(self, id):
        """Get a list of signals for this worker"""
        return self.get(f'/workers/{id}/signals')

    def worker_create_signal(self, id, execution_id, signal):
        """Create a signal for this worker"""
        return self.post(f'/workers/{id}/signals', data={
            'execution_id': execution_id,
            'signal': signal
        })
    
    def worker_deploy(self, number, batch, region, flavor, concurrency, prefetch=0):
        """Deploy (and create) some workers"""
        return self.put('/workers/deploy', data={
            'number':int(number),
            'region':region,
            'flavor':flavor,
            'batch':batch,
            'concurrency':concurrency,
            'prefetch': prefetch
        })
    
    def batch_stop(self,batch,signal=0):
        """Pause all workers for this batch"""
        if signal==0:
            return self.put(f'/batch/{batch}/stop',data={})
        return self.put(f'/batch/{batch}/stop',data={'signal':signal})
    
    def batch_go(self,batch,signal=0):
        """(re)set all workers affected to this batch to running"""
        return self.put(f'/batch/{batch}/go',data={'signal':signal})

    def batches(self):
        """List all batches, their tasks and workers"""
        return self.get(f'/batch/')

    def batch_delete(self,batch):
        """List all batches, their tasks and workers"""
        return self.delete(f'/batch/{batch}')

    def join(self, task_list, retry=1, check=False):
        """Wait for a certain list of tasks to succeed. In case of failure, relaunch tasks
        a limited number of time (retry). If check is True, then join will fail if
        one of the task fails (after all retries).
        Return a dictionary of the different status of tasks when tasks were all done
        """
        if not task_list:
            raise RuntimeError('task_list should not be empty')
        task_ids = [task['task_id'] if type(task)==dict else task.task_id for task in task_list]
        task_retries = dict([(task_id,0) for task_id in task_ids])

        all_task_done = False
        tasks = None
        dynamic_sleep_time = True
        sleep_time = JOIN_DYNAMIC_SLEEP_TIME_INCREMENT
        failed_tasks = []
        while not all_task_done:
            all_task_done = True
            old_tasks = tasks
            tasks = {'assigned':0,'failed':0, 'running':0, 'accepted':0, 'pending':0,'succeeded':0}
            for task in self.tasks():
                if type(task)==dict:
                    task = Namespace(**task)
                if task.task_id in task_ids:
                    if task.status=='failed':
                        if task_retries[task.task_id]<retry:
                            print(f'Retrying task {task.name or task.task_id} [{task_retries[task.task_id]+1}/{retry}]...')
                            self.task_update(task.task_id, status='pending')
                            task_retries[task.task_id]+=1
                            tasks['pending']+=1
                            all_task_done = False
                        else:
                            tasks['failed']+=1
                            if task.task_id not in failed_tasks:
                                print(f'Task {task.name or task.task_id} failed too many times giving up')
                                failed_tasks.append(task.task_id)
                    elif task.status in ['running','accepted','pending','assigned']:
                        all_task_done = False
                        tasks[task.status]+=1
                    else:
                        tasks[task.status]+=1
            print(f"Remaining tasks pending : {tasks['pending']}, assigned: {tasks['assigned']}, accepted: {tasks['accepted']}, running: {tasks['running']}, failed: {tasks['failed']}, succeeded: {tasks['succeeded']}")
            
            # sleeping if needed
            if not all_task_done:
                sleep(sleep_time)

            # adjust sleeping time
            if old_tasks is not None and dynamic_sleep_time:
                if tasks['succeeded']==old_tasks['succeeded'] and tasks['failed']==old_tasks['failed']:
                    # no task has been completed since last iteration so it is likely that the time loop is two short
                    sleep_time += JOIN_DYNAMIC_SLEEP_TIME_INCREMENT
                else:
                    # ok so now some tasks have been completed so it is likely the time loop is correct, no need to move it further
                    dynamic_sleep_time = False
            elif not dynamic_sleep_time and sleep_time>JOIN_DYNAMIC_SLEEP_TIME_INCREMENT:
                if tasks['succeeded']!=old_tasks['succeeded'] or tasks['failed']!=old_tasks['failed']:
                    # after all it seems it is moving quicker than expected, maybe time to shorten the time loop
                    dynamic_sleep_time = True
                    sleep_time -= JOIN_DYNAMIC_SLEEP_TIME_INCREMENT
        if check and tasks['failed']>0:
            raise RuntimeError('Could not complete all the tasks...')
        print('All tasks done!')
        return tasks


if __name__=="__main__":
    s=Server('127.0.0.1',style='object')



import boto3
import botocore.exceptions
from ftplib import FTP
import re
import logging as log
from functools import wraps, cached_property
from time import sleep
import shutil
import os
import requests
import glob
import hashlib
import threading
import subprocess
from .util import PropagatingThread
import concurrent.futures

# how many time do we retry
RETRY_TIME = 3
RETRY_SLEEP_TIME = 10
PUBLIC_RETRY_TIME = 20
MAX_PARALLEL_S3 = 5

class FetchError(Exception):
    pass

def pathjoin(*items):
    """Like os.path.join but always with '/' even with Windows and no double '/'
        in the middle of the path 
        (also no funny things if one of the item start with slash)
        (multiple slashes will be kept if at the begining or the end)"""
    last = len(items) - 1 
    joined_path = '/'.join([ item.rstrip('/') if rank==0 else 
                      item.lstrip('/') if rank==last else 
                      item.strip('/') 
                      for rank,item in enumerate(items)])
    # s3 tend to do funny things with that pattern, let us remove that buggy path
    joined_path = joined_path.replace('/./','/')
    return joined_path

# general
def retry_if_it_fails(n):
    """A decorator to retry n time some action"""
    def decorator(function):
        @wraps(function)
        def wrapper(*args, __retry_number__=None, **kwargs):
            iteration=0
            if __retry_number__ is None:
                __retry_number__ = n
            sleep_time=RETRY_SLEEP_TIME
            while iteration<__retry_number__:
                try:
                    retval = function(*args, **kwargs)
                    break
                except Exception:
                    log.exception('Something went bad')
                    iteration += 1
                    if iteration<__retry_number__:
                        log.warning(f'Waiting some time ({sleep_time}s)...')
                        sleep(sleep_time)
                        sleep_time *= 2
                        log.warning('Retrying...')
                    else:
                        raise
            return retval
        return wrapper
    return decorator    



def complete_if_ends_with_slash(source, destination):
    """This function checks if destination ends with slash in which case it completes
    with the source last item
    """
    if destination.endswith('/'):
        destination += source.split('/')[-1]
    return destination

# actions 

def gunzip(filepath):
    """Stupid gunzipper with gzip"""
    subprocess.run(['pigz','-d',filepath], check=True)
    
def untar(filepath):
    """Untar the tar archive locally where it is and delete the archive.
    (so it behaves like gunzip, and not like tar usually)"""
    path, basename = os.path.split(filepath)
    if basename.endswith('gz'):
        subprocess.run(['tar','xzf',basename], cwd=path, check=True)
    else:
        subprocess.run(['tar','xf',basename], cwd=path, check=True)
    os.remove(filepath)

# AWS S3 

S3_REGEXP=re.compile(r'^s3://(?P<bucket>[^/]*)/(?P<path>.*)$')

# this comes from here : https://github.com/boto/boto3/pull/2746
class BotoSession(boto3.session.Session):
    def client(self, *args, **kwargs):
        if kwargs.get('endpoint_url') is None and os.environ.get("AWS_ENDPOINT_URL"):
            kwargs['endpoint_url'] = os.environ.get("AWS_ENDPOINT_URL")
        return super().client(*args, **kwargs)

def get_s3():
    if os.environ.get("AWS_ENDPOINT_URL"):
        return boto3.resource('s3', 
            endpoint_url=os.environ.get("AWS_ENDPOINT_URL"))
    else:
        return boto3.resource('s3')

@retry_if_it_fails(RETRY_TIME)
def s3_get(source, destination):
    """S3 downloader: download source expressed as s3://bucket/path_to_file 
    to destination - a local file path"""
    log.warning(f'S3 downloading {source} to {destination}')
    destination=complete_if_ends_with_slash(source, destination)
    uri_match = S3_REGEXP.match(source).groupdict()
    try:
        if uri_match['path'].endswith('/'):
            bucket=get_s3().Bucket(uri_match['bucket'])
            jobs = {}
            with concurrent.futures.ThreadPoolExecutor(max_workers=MAX_PARALLEL_S3) as executor:
                for obj in bucket.objects.filter(Prefix=uri_match['path']):
                    destination_name = os.path.relpath(obj.key, uri_match['path'])
                    destination_name = os.path.join(destination, destination_name)
                    destination_path,_ = os.path.split(destination_name)
                    log.warning(f'S3 downloading {obj.key} to {destination_name}')
                    if not os.path.exists(destination_path):
                        os.makedirs(destination_path)
                    if not os.path.exists(destination_name):
                        jobs[executor.submit(bucket.download_file, obj.key, destination_name)]=obj.key
                for job in  concurrent.futures.as_completed(jobs):
                    obj = jobs[job]
                    log.warning(f'Done for {obj}: {job.result()}')
        else:
            BotoSession().client('s3').download_file(uri_match['bucket'],
                uri_match['path'],destination)
    except botocore.exceptions.ClientError as error:
        if 'Not Found' in error.response.get('Error',{}).get('Message',None):
            raise FetchError(f'{source} was not found')
        else:
            raise


@retry_if_it_fails(RETRY_TIME)
def s3_put(source, destination):
    """S3 uploader: download a local file path in source to a destination
    expressed as a s3 URI s3://bucket/path_to_file"""
    log.info(f'S3 uploading {source} to {destination}')
    destination=complete_if_ends_with_slash(source, destination)
    uri_match = S3_REGEXP.match(destination).groupdict()
    BotoSession().client('s3').upload_file(source,uri_match['bucket'],
        uri_match['path'])

# FTP

FTP_REGEXP=re.compile(r'^ftp://(?P<host>[^/]*)/(?P<path>.*)$')

@retry_if_it_fails(RETRY_TIME)
def ftp_get(source, destination):
    """FTP downloader: download source expressed as ftp://host/path_to_file 
    to destination - a local file path"""
    log.info(f'FTP downloading {source} to {destination}')
    destination=complete_if_ends_with_slash(source, destination)
    uri_match = FTP_REGEXP.match(source).groupdict()
    with open(destination, 'wb') as local_file:
        with FTP(uri_match['host']) as ftp:
            ftp.login()
            ftp.retrbinary(f"RETR {uri_match['path']}", local_file.write)

@retry_if_it_fails(RETRY_TIME)
def ftp_put(source, destination):
    """FTP uploader: upload source expressed as a local file path 
    to destination expressed as a FTP URL: ftp://host/path_to_file """
    log.info(f'FTP uploading {source} to {destination}')
    destination=complete_if_ends_with_slash(source, destination)
    uri_match = FTP_REGEXP.match(destination).groupdict()
    with open(destination, 'rb') as local_file:
        with FTP(uri_match['host']) as ftp:
            ftp.login()
            ftp.storbinary(f"STOR {uri_match['path']}", local_file.read)

@cached_property
def docker_available():
    """A property to see if docker is there"""
    return subprocess.run(['docker', '-v'],check=False).returncode==0

# aspera

ASPERA_REGEXP=re.compile(r'^fasp://(?P<url>.*)$')

@retry_if_it_fails(RETRY_TIME)
def fasp_get(source, destination):
    """Aspera downloader: download source expressed as fasp://user@host/path_to_file
    (NB usually with ENA, user is era-fasp) 
    to destination - a local file path"""
    log.info(f'Aspera downloading {source} to {destination}')
    if not docker_available:
        raise FetchError('Cannot use fasp (aspera) without docker')

    # we need destination to be expressed as a directory:
    # (but we remember that the target filename may be different from the source
    # filename)
    if destination.endswith('/'):
        target_filename = None
    else:
        destination, target_filename = os.path.split(destination)
    
    uri_match = ASPERA_REGEXP.match(source).groupdict()
    subprocess.run(f'docker run --rm -v {destination}:/output ibmcom/aspera-cli \
ascp -T --policy=high -l 300m -P33001 -m 30m -v \
-i /home/aspera/.aspera/cli/etc/asperaweb_id_dsa.openssh \
{uri_match["url"]} /output/', shell=True, check=True)
    
    # if the target filename is really different 
    if target_filename is not None:
        source_filename = uri_match['url'].split('/')[-1]
        if source_filename!=target_filename:
            shutil.move(os.path.join(destination, source_filename),
                os.path.join(destination,target_filename))

# run accessions containing fastq


FASTQ_RUN_REGEXP=re.compile(r'^run\+fastq://(?P<run_accession>[^/]*)/?$')

@retry_if_it_fails(RETRY_TIME)
def fastq_sra_get(run_accession, destination):
    """This subfunction of runacc_get is only called when EBI's ENA won't
    answer as NCBI's SRA while more complete is quite slow"""
    log.info(f'SRA get run+fastq://{run_accession} to {destination}')
    if not docker_available:
        raise FetchError('Cannot use SRA toolkit without docker')
    if not destination.endswith('/'):
            destination+='/'
    previous_fastq = glob.glob(os.path.join(destination, '*.fastq'))
    subprocess.run(f'docker run --rm -v {destination}:/destination ncbi/sra-tools \
        sh -c "cd /destination && prefetch {run_accession} && fasterq-dump -f --split-files {run_accession}"',
        shell=True,
        check=True)
    current_fastq = glob.glob(os.path.join(destination, '*.fastq'))
    fastqs = [fastq for fastq in current_fastq if fastq not in previous_fastq]
    subprocess.run(f'docker run --rm -v {destination}:{destination} ghcr.io/kasperskytte/docker-pigz:master -- '+
        ' '.join(fastqs),
        shell=True,
        check=True)

def _my_fastq_download(method, url, md5, destination):
    """A small adhoc function to download and check a fastq through a ftp_url plus a md5"""
    filename = url.split('/')[-1]
    if method=='fastq_ftp':
        ftp_get('ftp://'+url, destination, __retry_number__=1)
    elif method=='fastq_aspera':
        fasp_get(f'fasp://era-fasp@{url}', destination, __retry_number__=1)
    else:
        raise FetchError(f'No such method: {method}')
    with open(os.path.join(destination,filename),"rb") as f:
        readable_hash = hashlib.md5(f.read()).hexdigest()
        if readable_hash!=md5:
            raise FetchError(f'{filename} md5: {readable_hash} does not match ENA md5 {md5}')


@retry_if_it_fails(PUBLIC_RETRY_TIME)
def fastq_run_get(source, destination):
    """Fetch some fastq associated to a run accession"""
    log.info(f'Run accession: uploading {source} to {destination}')
    if not destination.endswith('/'):
            destination+='/'
    uri_match = FASTQ_RUN_REGEXP.match(source).groupdict()
    query_try = RETRY_TIME+1
    while query_try>0:
        try:
            run_query = requests.get(f"https://www.ebi.ac.uk/ena/portal/api/filereport?\
accession={uri_match['run_accession']}&result=read_run&fields=fastq_md5,fastq_aspera,\
fastq_ftp,sra_md5,sra_ftp&format=json&download=true&limit=0", timeout=30)
        except requests.Timeout:
            query_try -= 1
            continue
        break
    else:
        log.exception('EBI does not answer our query')
        return fastq_sra_get(uri_match['run_accession'], destination)
    if run_query.status_code==204:
        log.exception('This does not seem to be available on EBI')
        return fastq_sra_get(uri_match['run_accession'], destination)
    run = run_query.json()[0]



    if 'fastq_ftp' or 'fastq_aspera' in run:
        ftp_md5s = run['fastq_md5'].split(';')

        for method in ['fastq_ftp', 'sra', 'fastq_aspera']:
            if method == 'sra':
                return fastq_sra_get(uri_match['run_accession'], destination, 
                    __retry_number__=1)
            elif method in run:
                if method in ['fastq_aspera','sra'] and not docker_available:
                    continue
                urls =  run[method].split(';')

                # preparing to download and check all fastqs
                download_threads = []
                for url,md5 in zip(urls, ftp_md5s):
                    download_thread=PropagatingThread(target=_my_fastq_download,
                                        args=(method, url, md5, destination))
                    download_thread.start()
                    download_threads.append(download_thread)

                # waiting for all FTP to complete
                try:
                    for download_thread in download_threads:
                        download_thread.join()
                    return None
                except:
                    log.exception(f'EBI failed with method {method}')
                    continue
        raise FetchError(f'Could not fetch {source}')
    else:
        log.exception('EBI response does not include a fastq_ftp field')
        return fastq_sra_get(uri_match['run_accession'], destination)




# plain file

FILE_REGEXP=re.compile(r'^file://(?P<path>.*)$')

def file_get(source, destination):
    """A plain local copy from a file://... source to a local path
    really just some plain syntaxic sugar above shutil.copyfile"""
    log.info(f'FILE downloading {source} to {destination}')
    destination=complete_if_ends_with_slash(source, destination)
    uri_match = FILE_REGEXP.match(source).groupdict()
    shutil.copyfile(uri_match['path'], destination)

def file_put(source, destination):
    """Same as above except that source is this time a plain local path
    and destination is in the form file://... As above just some plain
    syntaxic sugar above shutil.copyfile """
    log.info(f'FILE uploading {source} to {destination}')
    destination=complete_if_ends_with_slash(source, destination)
    uri_match = FILE_REGEXP.match(destination).groupdict()
    if uri_match:
        complete_path = uri_match['path']
        path=os.path.dirname(complete_path)
        if not os.path.exists(path):
            os.makedirs(path)
        shutil.copyfile(source, complete_path)
    else:
        raise FetchError(f"Local URL did not match file://<path> pattern {destination}")


# generic wrapper

GENERIC_REGEXP=re.compile(r'^(?P<proto>[a-z0-9+]*)://(?P<resource>[^|]*)(\|(?P<action>.*))?$')

def get(uri, destination):
    """General downloader source should start with s3://... or ftp://...
    (source should not end with slash unless you know what you are doing if 
    destination ends with slash it will be completed with source end item)
    file://... is also supported but be careful that it is local to worker.
    """
    m = GENERIC_REGEXP.match(uri)
    if m:
        m = m.groupdict()
        source = f"{m['proto']}://{m['resource']}"
        if m['proto']=='s3':
            s3_get(source, destination)
        elif m['proto']=='ftp':
            ftp_get(source, destination)
        elif m['proto']=='file':
            file_get(source, destination) 
        elif m['proto']=='fasp':
            fasp_get(source, destination)
        elif m['proto']=='run+fastq':
            fastq_run_get(source, destination)       
        else:
            raise FetchError(f"This URI protocol is not supported: {m['proto']}")
        complete_destination = complete_if_ends_with_slash(source, destination)
        if m['action']=='gunzip':
            gunzip(complete_destination)
        elif m['action']=='untar':
            untar(complete_destination)
        elif m['action'] not in ['',None]:
            raise FetchError(f"Unsupported action: {m['action']}")
    else:
        raise FetchError(f'This URI is malformed: {uri}')

def put(source, uri):
    """General uploader destination should start with s3://.... or ftp://...
    (only anonymous ftp is implemented so put is unlikely to work with ftp)
    (source should not end with slash unless you know what you are doing if 
    destination ends with slash it will be completed with source end item)
    file://... is also supported but be careful that it is local to worker."""
    m = GENERIC_REGEXP.match(uri)
    if m:
        m = m.groupdict()
        if m['action']:
            raise FetchError(f'Action are unsupported when putting in {uri}')
        destination = f"{m['proto']}://{m['resource']}"
        destination = complete_if_ends_with_slash(source, destination)
        if m['proto']=='s3':
            s3_put(source, destination)
        elif m['proto']=='ftp':
            ftp_put(source, destination)
        elif m['proto']=='file':
            file_put(source, destination)        
        else:
            raise FetchError(f"This URI proto is not supported: {m['proto']}")
    else:
        raise FetchError(f'This URI is malformed: {uri}')

def check_uri(uri):
    """A small utility to check URI: return True if URI is valid, raise an exception otherwise"""
    m = GENERIC_REGEXP.match(uri)
    if m:
        m = m.groupdict()
        if m['proto'] not in ['ftp','file','s3','run+fastq']:
            raise FetchError(f"Unsupported protocol {m['proto']} in URI {uri}")
    else:
        raise FetchError(f"Malformed URI : {uri}")
    return True


import os

SQLALCHEMY_DATABASE_URI = os.environ.get('SQLALCHEMY_DATABASE_URI','sqlite:////tmp/scitq.db')
SQLALCHEMY_ECHO = False
SQLALCHEMY_TRACK_MODIFICATIONS = False
SECRET_KEY = '\xfb\x12\xdf\xa1@i\xd6>V\xc0\xbb\x8fp\x16#Z\x0b\x81\xeb\x16'
DEBUG = True
SQLALCHEMY_POOL_SIZE = os.environ.get('SQLALCHEMY_POOL_SIZE',None)


#!/usr/bin/env python
"""This script is designed to store and retrieve host, groups and hostvars in a 
sqlite database. It is simple but designed to be extremely performent
"""
import argparse
import sqlite3
import os
import json


# GLOBAL CONSTANTS
SQLITE_DATABASE = os.environ.get('ANSIBLE_SQLITE','/tmp/ansible-sqlite.db')
DEFAULT_GROUP = os.environ.get('ANSIBLE_DEFAULT_GROUP','Default')

# INTERNAL CONSTANTS
VERSION=1

class Database:
    def __init__(self, database_path):
        """Initialize database connection, if empty populate it"""
        self.connection = sqlite3.connect(database_path)
        self.connection.row_factory = sqlite3.Row
        self.base_cursor = self.connection.cursor()

        try:
            version = self.query('SELECT version FROM version')
            assert version[0][0]==VERSION
        except sqlite3.OperationalError:
            self.populate()

    def do(self, sql_query, values=[], nested=False):
        """Do a very basic query in an independant cursor"""
        if nested:
            c = self.connection.cursor()
        else:
            c = self.base_cursor
        c.execute(sql_query, values)
        return c

    def query(self, sql_query, values=[], nested=False):
        """Fetch all result of a basic query returning results - SELECT
        In case of nested query (avoid), nested must be set to true so that
        another cursor is claimed"""
        return self.do(sql_query, values, nested).fetchall()
    
    def populate(self):
        """Create the initial schema of the database"""
        self.do("""CREATE TABLE groups (
    group_id INTEGER NOT NULL,
    group_name VARCHAR(100) NOT NULL UNIQUE,
    PRIMARY KEY (group_id)
)""")
        self.do("""CREATE TABLE hosts (
    host_id INTEGER NOT NULL,
    host_name VARCHAR(100) NOT NULL UNIQUE,
    group_id INTEGER NOT NULL,
    PRIMARY KEY (host_id),
    FOREIGN KEY(group_id) REFERENCES groups (group_id)
)""")
        self.do("""CREATE TABLE hostvars (
    host_id INTEGER NOT NULL,
    variable VARCHAR(50) NOT NULL,
    value VARCHAR(100) NOT NULL,
    UNIQUE (host_id, variable)
)""")
        self.do("""CREATE TABLE version (
    version INTEGER NOT NULL
)""")
        self.do('INSERT INTO version (version) VALUES (?)', [VERSION])

        self.connection.commit()


    def list(self):
        """Main inventory function, list all hosts in all groups with _meta (hostvars)"""
        inventory = {}
        for group_name, host_name in self.query('SELECT group_name, host_name FROM hosts NATURAL JOIN groups'):
            if group_name not in inventory:
                inventory[group_name]=[]
            inventory[group_name].append(host_name)
        inventory['_meta'] = {}
        for host_name, variable, value in self.query('SELECT host_name, variable, value FROM hostvars NATURAL JOIN hosts'):
            if host_name not in inventory['_meta']:
                inventory['_meta'][host_name]={}
            inventory['_meta'][host_name][variable]=value
        return json.dumps(inventory,  indent=4, sort_keys=True)

    def list_host(self, host_name):
        """List all hostvars for a certain host, return an empty json if host has no vars or does not exists"""
        return json.dumps(dict(self.query('SELECT variable, value FROM hostvars NATURAL JOIN hosts WHERE host_name=?',
                                                (host_name,) )),
                        indent=4, sort_keys=True)

    def add_host(self, host_name, group_name):
        """Add a host in the inventory with default group_name - watch out: must
        be committed with self.connection.commit()"""
        ids = self.query('SELECT group_id FROM groups WHERE group_name=?',
                            (group_name,))
        if len(ids)==0:
            self.do('INSERT INTO groups (group_name) VALUES (?)',
                                (group_name,))
            ids = self.query('SELECT group_id FROM groups WHERE group_name=?',
                        (group_name,))   

        group_id = ids[0][0]

        try:
            self.do('INSERT INTO hosts (host_name, group_id) VALUES (?,?)',
                        (host_name, group_id))
        except sqlite3.IntegrityError:
            self.do('UPDATE hosts SET group_id=? WHERE host_name=?',(group_id,host_name))
                

    def get_host_id(self, host_name):
        """Return the host_id matching the host_name - this is safe because host_name are UNIQUE"""
        return self.query('SELECT host_id FROM hosts WHERE host_name=?',
                     (host_name,))[0][0]

    def set_hostvar(self, host_id, variable, value):
        """Insert a new host variable (or update) in table hostvars - value is converted to text"""
        try:
            self.do('INSERT INTO hostvars (host_id,variable,value) VALUES (?,?,?)',
                    (host_id, variable, value))
        except sqlite3.IntegrityError:
            self.do('UPDATE hostvars SET value=? WHERE host_id=? AND variable=?',
                    (value, host_id, variable))
    
    def del_host(self, host_name):
        """Delete a host and its hostvars in database"""
        host_id = self.get_host_id(host_name)
        self.do('DELETE FROM hostvars WHERE host_id=?',(host_id,))
        self.do('DELETE FROM hosts WHERE host_id=?',(host_id,))
        self.connection.commit()

def decorate_parser(parser):
    """Easing integration: all options to parser added here"""
    parser.add_argument('--list', default=None, action='store_true',
                help="List all hosts")
    parser.add_argument('--host', type=str, action='store', default=None,
                help="List one host in particular")
    parser.add_argument('--add-host', type=str, action='store', default=None,
                help=f"Add a host (in group {DEFAULT_GROUP} unless --in-group is used)")
    parser.add_argument('--in-group', type=str, action='store', default=DEFAULT_GROUP,
                help=f"Specify in which group a host should be added")
    parser.add_argument('--for-host', type=str, default=None,
                help="Specify a host where to add variables")
    parser.add_argument('--variable', type=str, action='append',
                help="Add a host variable (requires the host to be known, either with --add-host or with --for-host, and requires a value set with --value), can be used several times")
    parser.add_argument('--value', type=str, action='append',
                help="Add a host variable value (requires --variable, and thus the host to be known - with --add-host or --for-host), can be used several times")
    parser.add_argument('--del-host', type=str, action='store', default=None,
                help=f"Delete a host")

def inventory(args):
    """Main fonction"""
    
    if args.variable or args.value:
        if len(args.variable)!=len(args.value):
            raise (f"Use exactly the same number of values (value used: {args.value}) and variables (variables used: {args.variable})")
        if args.add_host is None and args.for_host is None:
            raise ("Cannot set some variables with values if host is unknown, use --for-host or --add-host")
        variables = dict(zip(args.variable, args.value))
    else:
        variables = {}

    db = Database(SQLITE_DATABASE)

    if args.list:
        return db.list()
    elif args.host is not None:
        return db.list_host(args.host)
    
    if args.add_host is not None:
        host = args.add_host
        db.add_host(host, args.in_group)
        if 'ansible_host' not in variables:
            variables['ansible_host']=host
    elif args.for_host is not None:
        host = args.for_host
    
    if variables:
        host_id = db.get_host_id(host)
        for variable, value in variables.items():
            db.set_hostvar(host_id, variable, value)
        db.connection.commit()

    if args.del_host is not None:
        db.del_host(args.del_host)
    

def scitq_inventory(list=False, host=None, add_host=None, in_group=DEFAULT_GROUP, 
        for_host=None, variable=[], value=[], del_host=None):
    """Same command as inventory but adapted for scitq internal use"""
    return inventory(argparse.Namespace(list=list, host=host, add_host=add_host, 
            in_group=in_group, for_host=for_host, variable=variable, 
            value=value, del_host=del_host))

if __name__=="__main__":
    parser = argparse.ArgumentParser(description='(yaf) Ansible SQLite inventory script')
    decorate_parser(parser)
    args = parser.parse_args()
    result = inventory(args)
    if result:
        print(result)


import threading
import os

class PropagatingThread(threading.Thread):
    """Taken from https://stackoverflow.com/questions/2829329/catch-a-threads-exception-in-the-caller-thread
    the join here will fails if one of the thread fails.
    """
    def run(self):
        """This version of run capture the exception occuring in the thread"""
        self.exc = None
        try:
            self.ret = self._target(*self._args, **self._kwargs)
        except BaseException as e:
            self.exc = e

    def join(self, timeout=None):
        """This version of join throw an exception if the thread joined did
        throw an exception"""
        super(PropagatingThread, self).join(timeout)
        if self.exc:
            raise self.exc
        return self.ret

    def is_alive(self):
        """This version of is_alive throw an exception if the thread ended
        with an exception"""
        if self.exc:
            raise self.exc
        return super(PropagatingThread, self).is_alive()

def package_path(*subdirs):
    """A very stupid hack to point to actual package data"""
    return os.path.join(os.path.dirname(__file__), *subdirs)


import argparse
from re import L
from .lib import Server, HTTPException
import socket
from time import sleep, time
import multiprocessing
import queue
import asyncio
from asyncio.subprocess import PIPE
import logging as log
import os
import psutil 
import platform
import queue
import tempfile
from .fetch import get,put,pathjoin
import traceback
import shutil
import subprocess
from signal import SIGTERM, SIGKILL, SIGTSTP, SIGCONT
import shlex

CPU_MAX_VALUE =10
POLLING_TIME = 4
READ_TIMEOUT = 5
QUEUE_SIZE_THRESHOLD = 2
IDLE_TIMEOUT = 600
DEFAULT_WORKER_STATUS = "running"
BASE_WORKDIR = os.environ.get("BASE_WORKDIR", 
    "/tmp" if platform.system() in ['Windows','Darwin'] else "/scratch")
BASE_RESOURCE_DIR = os.path.join(BASE_WORKDIR,'resource')

if not os.path.exists(BASE_RESOURCE_DIR):
    os.mkdir(BASE_RESOURCE_DIR)

DEFAULT_INPUT_DIR = '/input'
DEFAULT_OUTPUT_DIR = '/output'
DEFAULT_RESOURCE_DIR = '/resource'
DEFAULT_TEMP_DIR = '/tmp'
OUTPUT_LIMIT = 1024 * 128

def docker_command(input_dir, output_dir, temp_dir, resource_dir, cpu, extra_options,
        container, command):
    docker_cmd = ['docker','run', '--rm', '-d']
    if os.path.exists('/data'):
        docker_cmd.extend(['-v', '/data:/data'])
    docker_cmd.extend([
        '-v', f'{input_dir}:{DEFAULT_INPUT_DIR}',
        '-v', f'{output_dir}:{DEFAULT_OUTPUT_DIR}',
        '-v', f'{temp_dir}:{DEFAULT_TEMP_DIR}',
        '-v', f'{resource_dir}:{DEFAULT_RESOURCE_DIR}',
        '-e', f'CPU={str(cpu)}'])
    docker_cmd.extend(shlex.split(extra_options))
    docker_cmd.append(container)
    docker_cmd.extend(shlex.split(command)) 
    log.warning(f'Final docker command: {" ".join(docker_cmd)}')
    return docker_cmd

def create_dir(base_dir, sub_dir):
    """A small helper function to create a subdirectory and return its name.
    Add a final slash (or backslash with windows) to makes things clear it is a directory"""
    # os.path.join throw away the base dir
    if sub_dir.startswith(os.path.sep) or sub_dir.startswith('/'):
        sub_dir=sub_dir[1:]
    new_dir = os.path.join(base_dir, sub_dir)
    log.warning(f'Creating directory {new_dir}')
    os.mkdir(new_dir)
    # fix permission (this cannot be made with mode option in os.mkdir because 
    # it is restricted by umask which is not the case of os.chmod)
    os.chmod(new_dir, mode=0o777)
    if not new_dir.endswith(os.path.sep) and not new_dir.endswith('/'):
        new_dir += os.path.sep
    return new_dir

def final_slash(path):
    """Return the string with a final slash if it has not already"""
    return path if path.endswith('/') else path+'/'

def no_slash(path):
    """Do the reverse of final_slash, remove the final slash if it had one"""
    return path[:-1] if path.endswith('/') else path

class Executor:
    """Executor represent the process in which the task is launched, it is also
    designed to monitor the task and grab its output to push it regularly to the
    server.
    It performs now preparation tasks and post operation tasks."""

    def __init__(self, server, execution_id, task_id, command, input, output, 
                container, container_options, execution_started, execution_queue,
                cpu, resource_dir, resource, resources_db, run_slots, run_slots_semaphore,
                worker_id):
        self.s = Server(server, style='object')
        self.worker_id = worker_id
        self.execution_id = execution_id
        self.task_id = task_id
        self.command = command
        self.input = input
        if output and not output.endswith('/'):
            output+='/'
        self.output = output
        self.container = container
        self.container_options = container_options
        self.execution_started = execution_started
        self.execution_queue = execution_queue
        self.resource_dir = final_slash(resource_dir)
        self.workdir = tempfile.mkdtemp(dir=BASE_WORKDIR)
        log.warning(f'Workdir is {self.workdir} with base {BASE_WORKDIR}')
        self.input_dir = create_dir(self.workdir, DEFAULT_INPUT_DIR)
        self.output_dir = create_dir(self.workdir, DEFAULT_OUTPUT_DIR)
        self.temp_dir = create_dir(self.workdir, DEFAULT_TEMP_DIR)   
        self.cpu=cpu 
        self.resource=resource
        self.resources_db=resources_db
        self.run_slots=run_slots
        self.run_slots_semaphore=run_slots_semaphore
        self.container_id=None
        self.dynamic_read_timeout = READ_TIMEOUT
        asyncio.run(self.run())

    # via: https://stackoverflow.com/questions/10756383/timeout-on-subprocess-readline-in-python/34114767?noredirect=1#comment55978734_10756738
    async def execute(self, execution_id):
        # Start child process
        # NOTE: universal_newlines parameter is not supported
        log.warning(f'Run slots: {self.run_slots.value}')
        while self.run_slots.value<=0:
            log.warning(f'Overalocation, worker is out of run slot, have to wait...')
            self.run_slots_semaphore.release()
            sleep(POLLING_TIME)
            self.run_slots_semaphore.acquire()

        self.process = None
        if not self.container:
            try:
                self.process = await asyncio.create_subprocess_exec(
                        *shlex.split(self.command),
                        stdout=PIPE, stderr=PIPE, env={
                            'CPU':str(self.cpu),
                            'INPUT': no_slash(self.input_dir),
                            'OUTPUT': no_slash(self.output_dir),
                            'TEMP': no_slash(self.temp_dir),
                            'RESOURCE': no_slash(self.resource_dir)},
                        limit=OUTPUT_LIMIT)
                self.run_slots.value -= 1
                self.s.execution_update(execution_id, pid=self.process.pid, status='running')
            except Exception as e:
                self.s.execution_error_write(execution_id,
                        traceback.format_exc())
                self.s.execution_update(execution_id, status='failed')
        else:
            # this is the safe way to keep docker process attached while still getting its container id
            try:
                self.container_id = subprocess.run(docker_command(command=self.command, 
                        container=self.container,
                        input_dir=self.input_dir,
                        output_dir=self.output_dir,
                        temp_dir=self.temp_dir,
                        resource_dir=self.resource_dir,
                        cpu=self.cpu,
                        extra_options=self.container_options),
                    shell=False,
                    capture_output=True, check=True).stdout.decode('utf-8').strip()
                self.process = await asyncio.create_subprocess_exec(
                        'docker','attach',self.container_id,
                        stdout=PIPE, stderr=PIPE, limit=OUTPUT_LIMIT)
                self.run_slots.value -= 1
                self.s.execution_update(execution_id, pid=self.process.pid, status='running')
            except subprocess.CalledProcessError as e:
                self.s.execution_error_write(execution_id,e.stderr.decode('utf-8'))
                self.s.execution_update(execution_id, status='failed')
            except Exception as e:
                self.s.execution_error_write(execution_id,
                        traceback.format_exc())
                self.s.execution_update(execution_id, status='failed')

                




    async def get_output(self, execution_id):
        # Read line (sequence of bytes ending with b'\n') asynchronously
        output = []
        error = []

        now = time()
        while True:
            try:
                output.append(
                    (await asyncio.wait_for(self.process.stdout.readline(), self.dynamic_read_timeout)).decode('utf-8')
                )
                if time()-now>self.dynamic_read_timeout:
                    break
            except asyncio.TimeoutError:
                break
            except Exception as e:
                error.append(f'During stdout collection this error occured: {traceback.format_exc()}\n' )
                break
        self.s.execution_output_write(execution_id, ''.join(output))

        now = time()
        while True:
            try:
                error.append(
                    (await asyncio.wait_for(self.process.stderr.readline(), self.dynamic_read_timeout)).decode('utf-8')
                )
                if time()-now>self.dynamic_read_timeout:
                    break
            except asyncio.TimeoutError:
                break
            except Exception as e:
                error.append(f'During stderr collection this error occured: {traceback.format_exc()}\n' )
                break
        self.s.execution_error_write(execution_id,''.join(error))

        qsize = self.s.queue_size()
        if qsize > QUEUE_SIZE_THRESHOLD:
            self.dynamic_read_timeout += READ_TIMEOUT
        elif qsize == 0 and self.dynamic_read_timeout > READ_TIMEOUT:
            self.dynamic_read_timeout -= READ_TIMEOUT

        return self.process.returncode

    def download(self):
        """Do the downloading part, before launching, getting all input URIs into input_dir"""
        log.warning('Downloading input data...')
        if self.input:
            for data in self.input.split():
                get(data, self.input_dir)
        if self.resource:
            log.warning('Acquiring resources')
            log.warning(f'Resourcedb:  {self.resources_db}')
            for data in self.resource.split():
                if data not in self.resources_db or self.resources_db[data]=='failed':
                    self.resources_db[data]='lock'
                    try:
                        log.warning(f'Downloading resource {data}...')
                        get(data, self.resource_dir)
                        self.resources_db[data]='loaded'
                        log.warning(f'... resource {data} downloaded')
                    except:
                        self.resources_db[data]='failed'
                        log.warning(f'... resource {data} failed!')
                        raise
                else:
                    log.warning(f'Resource {data} is already there')
            while True:
                for data in self.resource.split():
                    if self.resources_db[data]=='lock':
                        log.warning(f'Waiting for resource {data}...')
                        sleep(POLLING_TIME)
                        break
                else:
                    log.warning(f'All resource seems there')
                    break
            for data in self.resource.split():
                if self.resources_db[data]=='failed':
                    self.resources_db[data]='lock'
                    try:
                        log.warning(f'Trying again to download resource {data}...')
                        get(data, self.resource_dir)
                        self.resources_db[data]='loaded'
                        log.warning(f'... resource {data} finally downloaded')
                    except:
                        self.resources_db[data]='failed'
                        log.warning(f'... resource {data} failed again!')
                        raise

    def upload(self):
        """Do the uploading part at the end, getting all output into output URI"""
        log.warning('Uploading output results...')
        if self.output:
            output_files = []
            for root, _, files in os.walk(self.output_dir):
                rel_path = os.path.relpath(root, self.output_dir)
                for local_data in files:
                    data = os.path.join(root, local_data)
                    put(data, pathjoin(self.output,rel_path,'/'))
                    output_files.append(local_data)
            if output_files:
                return ' '.join(output_files)
    
    def clean(self):
        """Clean working directory (triggered if all went well)"""
        shutil.rmtree(self.workdir)

    async def run(self):
        """Launching part of Executor
        """
        log.warning(f'Running job {self.execution_id}: {self.command}')
        self.execution_started.release()
        try:
            self.download()
        except Exception as e:
            self.s.execution_error_write(self.execution_id,
                traceback.format_exc())
            returncode=-1001
            self.s.execution_update(self.execution_id, 
                status='failed', 
                return_code=returncode, output='')
            return None
        
        self.run_slots_semaphore.acquire()
        while self.s.task_get(self.task_id).status == 'paused' or self.run_slots.value<=0:
            if self.run_slots.value<=0:
                log.warning(f'Task {self.task_id} has been prefetched and is waiting')
            else:
                log.warning(f'Task {self.task_id} is paused, waiting...')
            self.run_slots_semaphore.release()
            sleep(POLLING_TIME)
            self.run_slots_semaphore.acquire()
        # check a last time that this task is for us
        if self.s.execution_get(self.execution_id).status == 'failed':
            # the task is no longer for us so we bail out
            log.error(f'Execution {self.execution_id} was cancelled.')
            self.clean()
            return None
        
        log.warning(f'Launching job {self.execution_id}: {self.command}')
        await self.execute(execution_id=self.execution_id)
        self.run_slots_semaphore.release()

        if not self.process is None:
            log.warning('Launched')
            while True:
                log.warning(f"... {self.execution_id} ...")
                returncode = await self.get_output(self.execution_id)
                log.warning("...")
                if returncode is not None:
                    log.warning('... done')
                    await self.get_output(self.execution_id)
                    try:
                        output_files = self.upload()
                    except Exception as e:
                        output_files = ''
                        self.s.execution_error_write(self.execution_id,
                            traceback.format_exc())
                        returncode=-1000
                    if returncode!=0:
                        log.error(f'... task failed with error code {returncode}')
                    log.warning(f'Run slots: {self.run_slots.value}')

                    self.run_slots_semaphore.acquire()
                    self.s.execution_update(self.execution_id, 
                        status='succeeded' if returncode==0 else 'failed', 
                        return_code=returncode, output_files=output_files)
                    self.run_slots.value += 1
                    self.run_slots_semaphore.release()
                    if returncode==0:
                        self.clean()
                    break
                else:
                    try:
                        signal=self.execution_queue.get(block=False)
                        if self.container:
                            # sending a signal on the docker run process itself will not have any effect
                            # we must launch another process to send the signal to docker service instead
                            if self.container_id:
                                if signal == SIGTERM:
                                    subprocess.run(['docker','stop',self.container_id])
                                elif signal == SIGKILL:
                                    subprocess.run(['docker','kill',self.container_id])
                                elif signal == SIGTSTP:
                                    subprocess.run(['docker','pause',self.container_id])
                                elif signal == SIGCONT:
                                    subprocess.run(['docker','unpause',self.container_id])
                                else:
                                    log.warning(f'Cannot launch signal {signal} with docker...')
                        else:
                            self.process.send_signal(signal)
                        log.warning(f'...signal {signal} sent')    

                    except queue.Empty:
                        pass

class Client:
    def __init__(self, server, concurrency, name, batch):
        self.s = Server(server, style='object')
        self.server = server
        self.concurrency = concurrency
        self.prefetch = 0
        self.hostname = socket.gethostname()
        self.batch = batch
        if name is None:
            self.name = self.hostname
        else:
            self.name = name
        self.process = None
        self.executions = {}
        self.declare()
        self.has_worked = False
        self.idle_time = None
        self.resource_dir = tempfile.mkdtemp(dir=BASE_RESOURCE_DIR)
        self.manager = multiprocessing.Manager()
        self.resources_db = self.manager.dict()
        self.run_slots = multiprocessing.Value('i', concurrency)
        self.run_slots_semaphore = multiprocessing.Semaphore(1)


    def declare(self):
        """Declare a new worker or reconnect to the old one."""
        for worker in self.s.workers():
            if worker.name == self.name and worker.hostname == self.hostname:
                self.concurrency = worker.concurrency
                self.prefetch = worker.prefetch
                self.w=worker
                break
        else:
            self.w = self.s.worker_create(name=self.name,
                concurrency=self.concurrency,
                prefetch=self.prefetch,
                hostname=self.hostname,
                batch=self.batch)


    def run(self, status=DEFAULT_WORKER_STATUS):
        self.w = self.s.worker_update(self.w.worker_id, status=status)
        cpu_list=[]
        while True:
            try:
                try:
                    memory=round(100-psutil.virtual_memory()[2],1)
                    ##memory=round(VM[4]*100/VM[0],2) 
                    disk=psutil.disk_io_counters(perdisk=False, nowrap=True)
                    read_bytes,written_bytes = round(disk[2]/(1024**3),2),round(disk[3]/(1024**3),2)
                    #CPU1,CPU5,CPU15 = [round((x / psutil.cpu_count()) *100,2) for x in psutil.getloadavg()]
                    cpus = psutil.cpu_times_percent()
                    cpu_list.append(cpus.user)
                    if len(cpu_list)== CPU_MAX_VALUE:
                        del(cpu_list[0])
                    cpu_string = f'{cpus.user}'+ ('↑' if cpus.user >= sum(cpu_list)/len(cpu_list) else '↓')
                    if  platform.system() =='Linux':
                        cpu_string += f' / {cpus.iowait}'
                    log.warning(f'CPU is {cpu_string}')
                    self.w=self.s.worker_ping(self.w.worker_id, cpu_string, memory, read_bytes, written_bytes )
                    ##self.w=self.s.worker_ping(self.w.worker_id, load = ' '.join(map(
                        #lambda x: str(int(x*100)),psutil.getloadavg())))
                    ##
                except HTTPException as e:
                    if e.status_code==404:
                        self.declare()
                        continue
                if self.w.concurrency != self.concurrency:
                    self.run_slots_semaphore.acquire()
                    self.run_slots.value += self.w.concurrency-self.concurrency
                    self.run_slots_semaphore.release()
                    log.warning(f'Concurrency changed from {self.concurrency} to {self.w.concurrency}, adjusting run slots: {self.run_slots.value}')
                    self.concurrency = self.w.concurrency
                if self.prefetch != self.w.prefetch:
                    log.warning(f'Prefetch changed from {self.prefetch} to {self.w.prefetch}')
                    self.prefetch = self.w.prefetch
                if self.w.status=='running':
                    executions=self.s.worker_executions(self.w.worker_id, status='pending')
                    for execution in executions:
                        if execution.execution_id not in self.executions:
                            task = self.s.task_get(execution.task_id)
                            execution_started = multiprocessing.Semaphore(0)
                            execution_queue = multiprocessing.Queue()
                            p=multiprocessing.Process(target=Executor,
                                kwargs={
                                    'server': self.server,
                                    'execution_id': execution.execution_id,
                                    'task_id': task.task_id,
                                    'command': task.command,
                                    'input': task.input,
                                    'output': task.output,
                                    'execution_started': execution_started,
                                    'execution_queue': execution_queue,
                                    'container': task.container,
                                    'container_options': task.container_options,
                                    'cpu': psutil.cpu_count()//self.w.concurrency,
                                    'resource_dir': self.resource_dir,
                                    'resource': task.resource,
                                    'resources_db': self.resources_db,
                                    'run_slots': self.run_slots,
                                    'run_slots_semaphore': self.run_slots_semaphore,
                                    'worker_id': self.w.worker_id
                                })
                            self.s.execution_update(execution.execution_id, status='accepted')
                            p.start()
                            self.executions[execution.execution_id]=(p,execution_queue)
                            self.has_worked = True
                            self.idle_time = None
                            log.info('Waiting for execution to start')
                            execution_started.acquire()
                            log.info('Execution started')
                    if not(self.executions) and self.has_worked:
                        now = time()
                        if self.idle_time is None:
                            self.idle_time=now
                        elif now-self.idle_time > IDLE_TIMEOUT:
                            self.s.worker_callback(self.w.worker_id, message = "idle")
                            self.idle_time = None
                for signal in self.s.worker_signals(self.w.worker_id):
                    if signal.execution_id in self.executions:
                        log.warning(f'Sending signal {signal.signal} to execution {signal.execution_id}')
                        self.executions[signal.execution_id][1].put(signal.signal)
                    else:
                        log.warning(f'Execution {signal.execution_id} is not running in this worker')
                for execution_id in list(self.executions.keys()):
                    if not self.executions[execution_id][0].is_alive():
                        del(self.executions[execution_id])
                if self.run_slots_semaphore.acquire():
                    if self.concurrency-len(self.executions)!=self.run_slots.value:
                        log.warning(f'Race condition detected on process number')
                        # ok time to look what is really going on
                        running = waiting = 0
                        for execution in self.s.worker_executions(self.w.worker_id):
                            if execution.status=='running':
                                running += 1
                            elif execution.status=='accepted':
                                waiting += 1
                        log.warning(f'We are supposed to have {self.concurrency} running processes and we have {running}')
                        log.warning(f'We are supposed to have {self.prefetch} waiting processes and we have {waiting}')    
                        log.warning(f'Overall we should have {self.concurrency+self.prefetch} processes and we have {len(self.executions)}')
                        if self.run_slots.value != self.concurrency - running:
                            log.warning(f'Run slot derived, reseting from {self.run_slots.value} to {self.concurrency-running}')
                            self.run_slots.value = self.concurrency - running
                    self.run_slots_semaphore.release()
                else:
                    log.warning('Cannot estimate process number, giving up.')
            except Exception as e:
                log.exception('An exception occured during worker main loop')
            sleep(POLLING_TIME)
            
                
                    

def main():
    parser = argparse.ArgumentParser(description='scitq client')
    parser.add_argument('server', type=str,
                    help='IP address of the server')
    parser.add_argument('concurrency', type=int,
                    help='How many processes should be initially started')
    parser.add_argument('-n','--name', type=str, default=None,
                help="Give a name to the worker, default to hostname")
    parser.add_argument('-s','--status', type=str, default=DEFAULT_WORKER_STATUS,
            help=f"Give a name to the worker, default to hostname, default to {DEFAULT_WORKER_STATUS}")
    parser.add_argument('-b','--batch', type=str, default=None,
            help=f"Assign the worker to a default batch (default to None, the default batch)")
    args = parser.parse_args()
    
    Client(
        server=args.server, 
        concurrency=args.concurrency, 
        name=args.name,
        batch=args.batch
    ).run(status=args.status)
    

if __name__=="__main__":
    main()
    